import com.revusky.niggle.data.*;
import com.revusky.niggle.servlet.*;
import javax.servlet.http.*;
import java.io.IOException;
import java.util.List;


/**
 * The code that provides the dispatching
 * logic for our first pass at a rolodex
 */
public class MiniRoloServletInteraction extends ServletInteraction {

    /**
     * This constructor must have exactly this signature, since
     * Niggle invokes it by reflective lookup.
     */
    public MiniRoloServletInteraction(HttpServletRequest req, 
                                      HttpServletResponse resp, 
                                      NiggleConfig config) 
    throws IOException {
        super(req, resp, config);
    }

    public void execDefault() throws IOException
    {
        execEntries();
    }

    /**
     * Displays our default view, which is simply
     * a view of all the entries.
     */
    public void execEntries() throws IOException {
        page = getPage("entries.nhtml");
        List entries = getAllEntries();
        page.expose("entries", entries);
    }

    /**
     * deletes the record indicated by the unique_id
     * parameter and redirects to the default view.
     */
    public void execDelete() throws IOException {
        Integer recordID = new Integer(getParameter("unique_id"));
        Record entry = getEntry();
        getData().delete(entry);
        page = getPage("ack.nhtml");
        page.expose("deleted", true);
        page.expose("entry", entry);
    }

    /**
     * displays a page that allows
     * the user to edit the fields of a rolodex entry.
     */
    public void execEdit() throws IOException {
        page = getPage("edit.nhtml");
        page.expose("entry", getEntry());
    }

    /**
     * Processes the form embedded in the edit page template
     * and displays the acknowledgement page.
     */
    public void execProcess() throws IOException {
        boolean added = false;
        boolean modified = false;
        Record entry = getEntry();
        if (entry.isImmutable()) {
            Record newEntry = entry.getMutableCopy();
            DataUtil.fillRecordFromServletRequest(newEntry, this.request);
            getData().update(entry, newEntry);
            entry = newEntry;
            modified = true;
        }
        else {
            DataUtil.fillRecordFromServletRequest(entry, this.request);
            getData().insert(entry);
            added = true;
        }
        page = getPage("ack.nhtml");
        page.expose("added", added);
        page.expose("modified", modified);
        page.expose("entry", entry);
    }

    /**
     * a convenient hook for handling error conditions.
     */
    protected void recover(IOException e) throws IOException {
        try {
            throw e;
        }
        catch (DataException de) {
            Record entry = de.getRecord();
            if (entry != null) {
                page = getPage("edit.nhtml");
                page.expose("entry", entry);
            }
            else {
                page = getPage("ack.nhtml");
            }
            page.expose("error", de.getLocalizedMessage());
        }
    }

    protected void exposeDefaultVariables() throws IOException {
        super.exposeDefaultVariables();
        page.expose("defaultURL", getURL("default"));
        page.expose("editURL", getURL("edit"));
        page.expose("deleteURL", getURL("delete"));
        page.expose("processURL", getURL("process"));
    }

    /**
     * returns a handle to our one and only data source.
     */
    private MutableDataSource getData() throws IOException {
        DataRegistry reg = DataRegistryImpl.getDefaultRegistry();
        return(MutableDataSource) reg.getDataSource("rolodex_data");
    }

    /**
     * returns the rolodex entry based on the "unique_id" parameter
     * embedded in the servlet request. If the unique_id is not
     * defined, then the method returns a virgin entry.
     */
    private Record getEntry() throws IOException {
        try {
            Integer recordID = new Integer(getParameter("unique_id"));
            return getData().get(recordID);
        }
        catch (Exception e) {
            return newEntry();
        }
    }

    /**
     * returns a new, virgin instance of a rolodex entry.
     */
    private Record newEntry() throws IOException {
        DataRegistry reg = DataRegistryImpl.getDefaultRegistry();
        return reg.getExemplar("rolodex_entry");
    }

    /**
     * returns a java.util.List containing 
     * all the rolodex entries.
     */
    private List getAllEntries() throws IOException {
        MutableDataSource data = getData();
        return data.select(null); 
        // the null argument above implies we return 
        // all the entries in the data source, since otherwise,
        // we could pass in an instance of a RecordFilter object
        // that would allow us to retrieve a subset.
    }
}

