/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
 * Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * Neither the name of the Reminders On-Line nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

package com.revusky.niggle.data;

import java.io.IOException;

/**
 *  An interface implemented by objects that describe
 *  the constraints on the field of a niggle data record.
 *
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

public interface FieldDescriptor extends java.io.Serializable {
    /**
     *  A unique name that identifies the field
     */
    String getName();


    /**
     *  The kind of java object that can live
     *  in this record field.
     */
    Class getJavaType();


    /**
     * @return the name of this type in SQL
     * @throws UnsupportedOperationException
     */
    String getSQLType();


    /**
     * @return  the localized name of this field, used for
     *  localized error messages.
     */
    String getLocalizedName(java.util.Locale locale);


    /**
     *  Used in the core validity checks. Basically, this
     *  answers the question of whether a null
     *  (or the empty string in the case of a
     *  String field) is an acceptable value for the field.
     */
    boolean isRequired();


    /**
     *  Given the string representation of a field,
     *  return its internal java object representation.
     *  Used in deserializing from a flat file.
     */
    Object getObjectFromString(Record rec, String s) throws IOException;


    /**
     * returns the preferred the String representation
     *  of the object in question. Used to persist to
     *  a flat file.
     */
    String valueToString(Object o) throws IOException;


    /**
     *  Used when encapsulation requires Record::getFieldValue()
     *  to return a clone of an object. In the case of immutable
     *  objects such as java.lang.String, getCopy() can just
     *  return the object passed in as a parameter.
     * @return a copy of the object in question.
     */
    Object getCopy(Object o);


    /**
     *  "normalize" the string passed in. For example,
     *  with a case-insensitive string field, this method
     *  would return the lower case version of the string
     *  (or upper case if that is the "normal" form.)
     * @return the "normalized" string
     */
    String normalize(String s);


    /**
     *  Check whether this object is a valid value
     *  for this field of the given record. Used in the
     *  core validity checks.
     *
     * @exception  IOException  Description of Exception
     */
    void checkValidValue(Record rec, Object o) throws IOException;


    /**
     *  Does this FieldDescriptor describe a field of
     *  the Record? Useful for sanity checks.
     */
    boolean isFieldOf(Record rec);


    /**
     *  returns this field's "default value". By default,
     *  this will be null.
     */
    Object defaultValue();



    /**
     * Method to support slurping records from
     * a JDBC result set. Takes the result of a ResultSet.getObject()
     * and converts it to the JavaObject type we use.
     * In many cases, for strings and numbers, we should
     * get away with this just being a pass-through.
     */
    Object convertFromJDBC(Object obj);
}

