/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.data;

import java.io.*;
import java.util.Locale;
import java.util.Date;

/**
 * An interface that describes the canonical
 * methods of a data record in the niggle framework.
 * <P>
 * Basically, a Record should have the semantics of
 * an immutable set of values, where the values follow
 * a scheme given by the metadata in the associated RecordDescriptor
 * instance that can be got at via the method getMetadata()
 * <P>
 * The immutability semantics have a nuance.
 * A record's fields can be modified via set()
 * if the record is in an immutable state. Once a record is <em>frozen</EM>
 * its values are immutable. This typically occurs when the record is added to a
 * <CODE>MutableDataSource</CODE> container.
 * <P>
 * In this library, the way to "modify" data is via the
 * getMutableCopy() method. This will give you a copy that
 * can be modified. When that mutable copy is added to the
 * MutableDataSource, the original version will be deleted and
 * the new version will replace. It will then be initialized,
 * and thus, immutable.
 * <P>
 * The real virtue of this scheme is that when you are
 * working with a record that is immutable, you have
 * a pretty much iron-clad guarantee that all of the data in
 * the record is "valid" -- at least in the sense that it
 * matches the metadata description given by getMetadata().
 * You also have none of the typical synchronization worries,
 * where you have to think about the possibility that
 * another thread can change any values out from under you.
 *  <p>
 *  Each record has a unique ID which is set by
 *  the MutableDataSource when the record first put under
 *  management using <code>insertRecord()</code>.
 *  <p>
 * @see com.revusky.niggle.data.RecordDescriptor
 * @see com.revusky.niggle.data.DefaultRecord
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

public interface Record {
    /**
     * @return the name of this record type
     * subclasses must implement this. e.g. "reminder"
     */
    String getType();

    /**
     * a hook to be used by versioning schemes.
     */
    int getVersion();

    /**
     *  Retrieve the value of the record's primary key.
     *
     *  @return the value of this record's primary key, or null if it
     *  has not been set.
     */
    Object getPrimaryKey();

    /**
     *  set the value of this record's primary key.
     *  The record must be in a mutable state
     *  AND the primary key must be unset.
     */
    void setPrimaryKey(Object o);

    /**
     * create a clone.
     * A shallow copy of the record values
     * should be enough for immutable records
     * Note that a newly cloned object is
     * in a mutable state, so that you can modify it.
     */
    Record getMutableCopy();

    /**
     * Has this record been deleted or superseded in
     * the associated DataSource?
     */
    boolean isStale();

    /**
      * Have the fields all been set?
      * Once a record is put in a 
      * DataSource, its fields are immutable.
      */
    boolean isImmutable();

    /**
     * @return an object that describes the data held in this record.
     * @see com.revusky.niggle.data.RecordDescriptor
     */
    RecordDescriptor getMetadata();

    /**
     * Low-level method to query the value of a field in a Record.
     * concrete implementations will probably wrap this.
     * @param field to query.
     * @return an Object wrapper around field value.
     * @deprecated still exists for backward compatibility
     */
    Object get(FieldDescriptor field);

    /**
     * Low-level method to set an individual field value
     * concrete implementations will probably wrap this.
     * @param field to set.
     * @param value Object wrapping the value
     * @throws InvalidFieldException if there is no field of that name.
     * @throws ImmutableDataException if this record
     * is immutable @see #getMutableCopy()
     */
    void set(FieldDescriptor field, Object value);


    /**
     * @return a Locale for displaying the record.
     */
    Locale getDisplayLocale();


    /**
     * reset the fields to their default state
     * The record must be in a mutable state.
     * @throws ImmutableDataException if this record
     * is immutable @see #getMutableCopy()
     */
    void clearFields();

    /**
     * Low-level method to set the value of a field.
     * Concrete implementations will probably wrap this
     * with a higher-level get/set API.
     * In theory, this method breaks encapsulation
     * and should only be called internally from concrete
     * implementing classes. In practice, outside use may sometimes
     * be justified by efficiency considerations.
     * @throws ImmutableDataException if this record
     * is immutable @see #getMutableCopy()
     */
    void set(int i, Object value);

    /**
     * Low-level method to set the value of a field.
     * Concrete implementations will probably wrap this
     * with a higher-level get/set API.
     * @param fieldname the name of the field to set.
     * @throws InvalidFieldException if there is no field of that name.
     * @throws ImmutableDataException if this record
     * is immutable @see #getMutableCopy()
     */
    void set(String fieldname, Object value);

    /**
     * Low-level method to get the value of a field.
     * Concrete implementations will probably wrap this
     * with a higher-level get/set API.
     * In theory, this method breaks encapsulation
     * and should only be called internally from concrete
     * implementing classes. In practice, outside use might sometimes
     * be justified by efficiency considerations.
     */
    Object get(int i);

    /**
     * Low-level method to set the value of a field.
     * Concrete implementations will probably wrap this
     * with a higher-level get/set API.
     * @param fieldname
     * @throws InvalidFieldException if there is no field of that name.
     */
    Object get(String fieldname);

    /**
     * convenience method that returns the value of the field
     * that holds the record's creation timestamp.
     * This could conceivably throw an UnsupportedOperationException
     * if the object has no creation_date field.
     */
    Date getCreationDate();

    /**
     * convenience method that returns the value of the field
     * that holds the record's last modified timestamp.
     * This could conceivably throw an UnsupportedOperationException
     * if the object has no last_modified field.
     */
    Date getLastModified();

    /**
     * mark this record as having just been modified.
     * This could conceivably throw an UnspoortedOperationException
     * if the object has no last_modified field.
     * Also, the record must be in a mutable state.
     */
    void touch();

    /**
     * Method only used internally by niggle.
     */
    void setMetadata(RecordDescriptor desc);

    /**
     * The special reserved name of the type key.
     */
    String TYPE_KEY = "_meta";

    String STATUS_KEY = "_status";

    /**
     * By convention, the integral value of the id
     * of a record whose unique_id is unset.
     */
    int INVALID_ID = Integer.MIN_VALUE;
}

