/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.data;

import java.io.*;
import com.revusky.niggle.util.*;

/**
 * An interface for objects that can filter a set of records
 * based on some criterion. A unary predicate that operates on
 * records, more formally. 
 * A RecordFilter has a method called accept() that allows
 * us to filter over a set of in-memory records. The 
 * sqlEquivalent() should return the WHERE clause
 * that would provide the same subset of records from an 
 * RDBMS with an SQL interface.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

public interface RecordFilter extends Serializable {
    /**
     * The method that determines whether the
     * record gets through the filter.
     */
    public boolean accept(Record rec);

    /**
     * Method that returns the SQL WHERE clause that
     * is the equivalent of the unary predicate that
     * this field represents. It is the programmer's
     * responsibility that this really be equivalent.
     * For programmer convenience, the string returned
     * should start with a space and then WHERE
     */
    public String sqlEquivalent();

    public RecordFilter NULL_FILTER = new NullFilter();


    public class NullFilter implements RecordFilter {
        public boolean accept(Record rec) {
            return rec != null;
        }

        public String sqlEquivalent() {
            return "";
        }
    }

    public class ClassConstraint implements RecordFilter {
        public final Class cl;
        public ClassConstraint(Class cl) {
            this.cl = cl;
            if (!cl.isInterface())
                Assert.assert(Record.class.isAssignableFrom(cl),
                              "You should specify a Record subtype");
        }

        public boolean accept(Record rec) {
            if (rec instanceof RecordReference) {
                rec = ((RecordReference) rec).getUnderlyingRecord();
            }
            return(rec!= null) && cl.isInstance(rec);
        }

        public String sqlEquivalent() {
            throw new UnsupportedOperationException();
        }
    }

    public class FieldConstraint implements RecordFilter {
        public final String fieldName;
        public final Object value;
        private String whereClause;

        public FieldConstraint(String fieldName, Object value) {
            this.fieldName = fieldName;
            if ("".equals(value))
                value = null;
            this.value = value;
            String s = value != null ? value.toString() : null;
            if (value instanceof String) {
                s = "'" + value + "'";
            }
            if (value != null) 
                whereClause = " WHERE " + fieldName + "=" + s;
            else
                whereClause = " WHERE " + fieldName + " is NULL";
        }

        public boolean accept(Record rec) {
            if (rec ==null)
                return false;
            if (value == null) {
                Object value = rec.get(fieldName);
                return value == null || value.equals("");
            }
            return value.equals(rec.get(fieldName));
        }

        public String sqlEquivalent() {
            return whereClause;
        }
    }

    public class TypeConstraint implements RecordFilter {
        public final String type;

        /**
         * @param type The type of record to accept
         */
        public TypeConstraint(String type) {
            this.type = type;
        }

        public boolean accept(Record rec) {
            return(rec != null) && rec.getType().equals(type);
        }

        public String sqlEquivalent() {
            String msg = "This constraint makes no sense in SQL, "
                         + "since all the rows returned will necessarily be of a given type.";
            throw new UnsupportedOperationException(msg);
        }
    }

    public class UnionFilter implements RecordFilter {
        final RecordFilter first, second;

        public UnionFilter(RecordFilter first, RecordFilter second) {
            this.first = first;
            this.second = second;
        }

        public boolean accept(Record rec) {
            return first.accept(rec) || second.accept(rec);
        }

        public String sqlEquivalent() {
            String clause1 = first.sqlEquivalent();
            String clause2 = second.sqlEquivalent();
            if (!Util.hasContent(clause1))
                return clause2;
            if (!Util.hasContent(clause2))
                return clause1;
            clause1 = clause1.trim().substring("WHERE ".length());
            clause2 = clause2.trim().substring("WHERE ".length());
            return " WHERE (" + clause1
            + ") OR ("
            + clause2
            + ")";
        }
    }

    public class IntersectionFilter implements RecordFilter {
        final RecordFilter first, second;

        public IntersectionFilter(RecordFilter first, RecordFilter second) {
            this.first = first;
            this.second = second;
        }

        public boolean accept(Record rec) {
            return first.accept(rec) && second.accept(rec);
        }

        public String sqlEquivalent() {
            String clause1 = first.sqlEquivalent();
            String clause2 = second.sqlEquivalent();
            if (!Util.hasContent(clause1))
                return clause2;
            if (!Util.hasContent(clause2))
                return clause1;
            clause1 = clause1.trim().substring("WHERE ".length());
            clause2 = clause2.trim().substring("WHERE ".length());
            return " WHERE (" + clause1
            + ") AND ("
            + clause2
            + ")";
        }
    }

    public class NegationFilter implements RecordFilter {
        final RecordFilter filter;

        public NegationFilter(RecordFilter filter) {
            this.filter = filter;
        }

        public boolean accept(Record rec) {
            return !filter.accept(rec);
        }

        public String sqlEquivalent() {
            String clause = filter.sqlEquivalent();
            if (!clause.toUpperCase().trim().startsWith("WHERE"))
                throw new UnsupportedOperationException();
            clause = "(" + clause.trim().substring("WHERE ".length()) + ")";
            return " WHERE NOT " + clause;
        }
    }
}
