/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.data.inmemory;

import com.revusky.niggle.data.*;
import java.io.*;
import java.util.*;
import com.revusky.niggle.util.*;

/**
 * A data listener that persists using an extremely
 * crude scheme. When a record is added,
 * it appends it to the flat file. When a record is deleted,
 * it appends an AntiRecord (dummy record)
 * to a flat file.
 * Periodically, it writes the entire flat file from scratch.
 * At that time, obsolete records are not written effectively purging all
 * knowlege of their existence.
 * This can be a decent first pass while working
 * up an application, but is not scaleable, of course.
 * The flat files use a readable ASCII format, not the usual
 * Java serialised object format.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
public class CrudeDataPersister implements DataListener {

    /**
     * DataSource we listen to, to find out about records to persist.
     */
    final MutableDataSource dataSource;

    /**
     * Flat file where we persist records
     */
    final String filename;

    /**
     * How many records we have added without doing a complete rewrite of the persist file.
     */
    private int persistCount;

    /**
     * How often to do a complete rewrite of the persist file,
     */
    private final int persistFrequency;

    /**
     * This constructor automatically
     * adds this object to the list of listeners
     * of the specified DataSource
     * @param dataSource the DataSource whose events to
     * listen for and persist.
     * @param filename the flat file to persist to.
     */
    public CrudeDataPersister(MutableDataSource dataSource, String filename, int persistFrequency)
    throws IOException {
        Assert.canWriteToFile(filename);
        this.filename = filename;
        this.dataSource = dataSource;
        this.persistFrequency = persistFrequency;
        //Start at persistFrequency so that we start off with a persist.
        this.persistCount = persistFrequency;
        dataSource.addDataListener(this);
    }

    public void recordDeleted(DataEvent event) throws IOException {
        appendRecord(new AntiRecord(event.oldRecord));
    }


    public void recordUpdated(DataEvent event) throws IOException {
        writeLine(DefaultRecordInputStream.UPDATE_MARKER);
        writeRecord(event.newRecord);
    }

    public void recordInserted(DataEvent event) throws IOException {
        writeLine(DefaultRecordInputStream.INSERTION_MARKER);
        writeRecord(event.newRecord);
    }

    /**
     * writes a record to the output file,
     * usually appends, but occasionally compacts,
     * with frequency based on persistFrequency
     */
    private void writeRecord(Record rec) throws IOException {
        synchronized (dataSource) {
            ++persistCount;
            if (persistCount > persistFrequency) {
                persistRecords();
                persistCount = 0;
            }
            else {
                appendRecord(rec);
            }
        }
    }

    /**
     * append a record to the flat file.
     */
    private void appendRecord(Record rec) throws IOException {
        FileOutputStream fos = new FileOutputStream(filename, true);
        ObjectOutput out = new DefaultRecordOutputStream(fos);
        out.writeObject(rec);
        out.flush();
        out.close();
    }

    /**
     * Totally rewrite the persist flat file from scratch.
     * Obsolete and deleted records will
     * not be written, but every other record will be.
     */
    private void persistRecords() throws IOException {
        String tempFilename = filename + ".###";
        ObjectOutput out = new DefaultRecordOutputStream(tempFilename);
        Iterator it = dataSource.select(null).iterator();
        while (it.hasNext()) {
            out.writeObject(it.next());
        }
        out.flush();
        out.close();
        File tempFile = new File(tempFilename);
        File prevFile = new File(filename);
        FileLock lock = new FileLock(filename);
        prevFile.delete();
        tempFile.renameTo(prevFile);
        lock.dispose();
    }

    /**
     * writes a line to the flat file
     * (appends the \n to the string)
     */
    private void writeLine(String s) throws IOException {
        PrintWriter pw = new PrintWriter(new FileWriter(filename, true), true);
        pw.println(s);
        pw.close();
    }
}
