/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.data.inmemory;

import java.util.*;
import java.io.IOException;
import com.revusky.niggle.data.*;
import com.revusky.niggle.util.*;

/**
 * An object that listens to the insert/update/delete events of a MutableDataSource
 * and maintains a lookup table based on a certain field.
 * It works for both one-to-one and one-to-many lookups
 * The primary record manager does lookup by the uniqueID.  This object
 * does lookups by other fields such as username.
 * MemoryDataSource finds out about changes via the DataListener interface.
 * There is one MemoryDataSource per field indexed.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
public class InMemoryIndex extends AbstractDataSource implements DataListener {

    /**
     * The actual container used to hold the info.
     */
    private HashMap cache = new HashMap();

    /**
     * lookup field
     */
    private String field;


    /**
     * The records to index. If this is null,
     * we just try to index them all.
     */
    private RecordFilter filter;

    private String name;

    /**
     * true if we can have multiple records with the same key.
     */
    private boolean multipleAllowed;

    synchronized public void recordUpdated(DataEvent event) throws IOException {
        recordDeleted(event);
        recordInserted(event);
    }

    /**
     * Listener notified when a record has been deleted.
     * @param event describes the record deleted.
     */
    synchronized public void recordDeleted(DataEvent event) throws IOException {
        if ( filter != null && !filter.accept(event.oldRecord) )
            return;
        Object key = null;
        try {
            key = event.oldRecord.get(field);
            // Allow for non-required fields.
            if ((key == null) || "".equals(key))
                return;
        }
        catch ( InvalidFieldException e ) {
            // If the field is invalid on this record, we just ignore.
            return;
        }
        Object value = cache.get(key);
        if ( value == null ) {
            Logger.log("Tried to remove record from cache with keyField "
                       + field
                       + " of value "
                       + key
                       + " but it was not in the cache.",
                       Logger.LOG_WARN);
            return;
        }
        if ( !multipleAllowed ) {
            cache.remove(key);
        }
        else {
            if ( value instanceof List ) {
                List l = (List) value;
                if ( !l.remove(event.oldRecord) )
                    Logger.log("Tried to remove record from cache with keyField "
                               + field
                               + " of value "
                               + key
                               + " but it was not in the cache.",
                               Logger.LOG_WARN);
                if ( l.size() == 0 )
                    cache.remove(key);
                else if ( l.size() == 1 )
                    cache.put(key, l.get(0));
            }
            else {
                cache.remove(key);
            }
        }
    }


    /**
     * add a record, and index it.
     * @param newRec the record to add.
     */
    protected void cacheRecord(Record newRec) {
        if ( filter != null && !filter.accept(newRec) )
            return;

        Object keyValue = null;
        try {
            keyValue = newRec.get(field);
        }
        catch ( InvalidFieldException e ) {
            // If this record has no field of the indicated key, then we just
            // don't cache it. Allows us to listen to events on heterogeneous
            // MutableDataSources and simply ignore the irrelevant events.
            return;
        }

        // Allow for non-required fields.
        if ((keyValue == null) || "".equals(keyValue))
            return;

        Object currentValue = cache.get(keyValue);
        if ( currentValue == null ) { // no problem here.
            cache.put(keyValue, newRec);
            return;
        }
        if ( !multipleAllowed ) {
            Record oldRec = (Record) currentValue;
            if (!oldRec.isStale()) {
                Logger.log("Adding record:\n "
                           + newRec
                           + " to the cache, displacing the older record:\n "
                           + oldRec,
                           Logger.LOG_WARN);
            }
            cache.put(keyValue, newRec);
        }
        else if ( currentValue instanceof List ) {
            ((List) currentValue).add(newRec);
        }
        else {
            List l = new ArrayList();
            l.add(currentValue);
            l.add(newRec);
            cache.put(keyValue, l);
        }
    }


    /**
      * Listener notified when a record has been added.
      * @param event describes the record added.
      */
    synchronized public void recordInserted(DataEvent event) throws IOException {
        cacheRecord(event.newRecord);
    }

    /**
     *  Return the keys to the index.
     *
     *  @return a copy of the keys to the index.  The list can be empty,
     *  but never null.
     *
     *  @exception IOException thrown in case of a low-level error.
     */
    synchronized public List keys(String type) throws IOException {
        if (Util.hasContent(type))
            throw new UnsupportedOperationException();
        return new ArrayList(cache.keySet());
    }

    synchronized public Record get(String type, Object key) throws IOException {
        Object value = cache.get(key);
        if (value == null && key instanceof String) {
            // This is a kludge from hell! :-) (JR)
            // We check whether the application programmer
            // really meant to pass in an int instead of the String!!!
            try {
                value = cache.get(new Integer(key.toString()));
            }
            catch (Exception e) {
            }
        }
        if (value == null || value instanceof Record)
            return(Record) value;
        else if (value instanceof List) {
            List l = (List) value;

            if (l.isEmpty()) {
                cache.remove(key);
                throw new MissingRecordException("No record for key " +
                                                 key.toString());
            }
            for (Iterator it = l.iterator(); it.hasNext();) {
                Record rec = (Record) it.next();
                if (type == null || type.equals(rec.getType()))
                    return rec;
            }
        }
        return null;
    }

    /**
     *  Use this routine to lookup if there may be multiple records with
     *  the same key.
     *
     *  @param key the key on which the record is indexed.
     *
     *  @return a List of matches (which may be empty)
     */
    synchronized public List getRecords(Object key) throws IOException {
        List result = new ArrayList();
        Object value = cache.get(key);
        if (value == null && key instanceof String) {
            // kludge here to deal with strings that are supposed
            // to be ints.
            try {
                value = cache.get(new Integer(key.toString()));
            }
            catch (Exception e) {
            }
        }
        if (value instanceof List)
            result.addAll((List) value);
        else {
            if (value != null)
                result.add(value);
        }
        return result;
    }

    public String getName() {
        return name;
    }


    /**
     * Called to run over a MutableDataSource's entire set of elements
     * and index them.
     * @param rm primary record manager that contains the records.
     */
    public void attach(MutableDataSource dataSource) throws IOException {
        dataSource.addDataListener(this); // just in case anything happens meanwhile.
        Iterator it = dataSource.select(filter).iterator();
        while ( it.hasNext() ) {
            cacheRecord((Record) it.next());
        }
    }

    /**
     * clear the cache -- wipe out the entire index structure.
     */
    public void clear() throws IOException {
        cache.clear();
    }

    /**
     * Constructor
     * Create an MemoryDataSource that caches on the given keyField
     * If type is non-null, it will ignore all records that are not of this type
     * @param field the name of the field to use as the lookup key
     * @param type the object type to index. If this is null, we just try to index everything.
     * @param multipleAllowed determines whether we maintain a one-to-many or one-to-one cache.
     */
    public InMemoryIndex(String name, String field, RecordFilter filter, boolean multipleAllowed)
    throws IOException {
        this.filter = filter;
        this.field = field;
        this.multipleAllowed = multipleAllowed;
        this.name = name;
    }

    /**
     * A convenience constructor largely for backward compatibility.
     */
    public InMemoryIndex(String name, String field, Class cl, boolean multipleAllowed)
    throws IOException {
        this(name, field, (RecordFilter) null, multipleAllowed);
        if (cl != null)
            this.filter = new RecordFilter.ClassConstraint(cl);
    }

    public InMemoryIndex() throws IOException {}

    public void init(Properties props) throws IOException {
        this.name = props.getProperty("NAME");
        this.field = props.getProperty("INDEX_FIELD");
        String recManagerKey = props.getProperty("PRIMARY_CONTAINER");
        String multiple = Util.noNull(props.getProperty("MULTIPLE_ALLOWED"));
        String className = props.getProperty("CLASS_CONSTRAINT");
        String typeName = props.getProperty("TYPE_CONSTRAINT");
        Assert.assert(Util.hasContent(field), "Must specify index field");
        Assert.assert(!Util.hasContent(className) || !Util.hasContent(typeName), "cannot specify both a type and class constraint");
        MutableDataSource rm = (MutableDataSource) DataRegistryImpl.getDefaultRegistry().getDataSource(recManagerKey);
        Assert.assert(rm != null, this.name + " need " + recManagerKey);
        this.multipleAllowed = multiple.startsWith("y") || multiple.startsWith("Y") || multiple.startsWith("t") || multiple.startsWith("T");
        if (Util.hasContent(className)) {
            try {
                Class cl = Class.forName(className);
                this.filter = new RecordFilter.ClassConstraint(cl);
            }
            catch (Exception e) {
                Assert.rethrow(e);
            }
        }
        else if (Util.hasContent(typeName)) {
            this.filter = new RecordFilter.TypeConstraint(typeName);
        }
        rm.addDataListener(this);
        Iterator it = rm.select(filter).iterator();
        while ( it.hasNext() ) {
            cacheRecord((Record) it.next());
        }
    }

    public void close() throws IOException {}

}

