/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (C) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.data.jdbc;

import java.io.*;
import java.sql.*;
import java.util.*;
import com.revusky.niggle.data.*;
import com.revusky.niggle.util.*;

/**
 * A class that is essentially a facade in front of an external
 * relational database table. This is just one of various
 * possible implementations. This implementation allows for
 * an in-memory cache of adjustable size. This container
 * is probably highly efficient for a web app whose profile is that
 * there are a lot of reads and relatively few writes to the 
 * external RDBMS. That is probably more the rule than the exception.
 * @author <A HREF="mailto:jon@revusky.com">Jonathan Revusky</A>
 */

public class JDBCBackedRecordSet extends AbstractMutableDataSource implements DataListener {

    String driverClassName, jdbcURL, defaultRecordType;

    // The maximum size of the cache.
    int maxCacheSize;

    // If all the records of a given type are cached,
    // we can avoid going to the RDBMS.
    private List completelyCachedTypes = new ArrayList();

    public JDBCBackedRecordSet() throws IOException
    {
    }

    public void recordDeleted(DataEvent event) throws IOException {
        try {
            getConnection().performDeletion(event.oldRecord);
            uncacheRec(event.oldRecord);
        }
        catch (Exception e) {
            Logger.logException(e);
        }
    }

    public void recordInserted(DataEvent event) throws IOException {
        insert(event.newRecord);
    }

    public void recordUpdated(DataEvent event) throws IOException {
        update(event.oldRecord, event.newRecord);
    }

    public void init(Properties props) throws IOException {
        super.init(props);
        this.driverClassName = props.getProperty("JDBC_DRIVER_CLASS");
        this.jdbcURL = props.getProperty("JDBC_URL");
        this.defaultRecordType = props.getProperty("DEFAULT_RECORD_TYPE");
        String cacheSizeString = props.getProperty("MAX_CACHE_SIZE");
        try {
            Class.forName(driverClassName);
            if (Util.hasContent(cacheSizeString))
                maxCacheSize = Integer.parseInt(cacheSizeString);
        }
        catch (Exception e) {
            Assert.rethrow(e);
        }
        String recNames = props.getProperty("CREATE_TABLES_IF_NECESSARY");
        StringTokenizer st = new StringTokenizer(recNames, " ,\n\r\t");
        try {
            Connection conn = DriverManager.getConnection(jdbcURL);
            List existingTables = SQLUtil.getTableNames(conn);
            while (st.hasMoreTokens()) {
                String recName = st.nextToken();
                RecordDescriptor metadata = DataRegistryImpl.getDefaultRegistry().getMetadata(recName);
                String tableName = metadata.getTableName();
                if (!existingTables.contains(tableName.toLowerCase()))
                    SQLUtil.createTable(conn, metadata);
            }
            conn.close();
        }
        catch (SQLException e) {
            Logger.logException(e);
            throw new DataException(null, e);
        }
    }

    synchronized public void delete(Record rec) throws IOException {
        try {
            getConnection().performDeletion(rec);
        }
        catch (SQLException se) {
            throw new DataException(rec, se);
        }
        markStale(rec);
        uncacheRec(rec);
        fireDataEvent(new DataEvent(this, DataEvent.RECORD_DELETED, null, rec));
    }

    synchronized public void insert(Record rec) throws IOException {
        Object key = rec.getPrimaryKey();
        if ("".equals(key))
            key = null;
        if (key == null) {
            rec.setPrimaryKey(uniqueKey(rec));
        }
        freeze(rec);
        try {
            getConnection().performInsertion(rec);
        }
        catch (SQLException se) {
            throw new DataException(rec, se);
        }
        cacheRec(rec);
        DataEvent de = new DataEvent(this, DataEvent.RECORD_INSERTED, rec, null);
        fireDataEvent(de);
    }

    synchronized public void update(Record oldRec, Record newRec) throws IOException {
        Assert.assert(!newRec.isImmutable(),
                      "Expecting a mutable record.");
        Assert.assert(oldRec.isImmutable(), "You can only update a record that is already initialized.");
        Assert.assert(!newRec.isStale(),
                      "This record is stale. " +
                      "You cannot add it to the data source.");
        if (oldRec.isStale()) {
            Record currentRec = get(oldRec.getType(), oldRec.getPrimaryKey());
            throw new ModifiedRecordException(currentRec, "attempt to update a version that is not current");
        }
        Object primaryKey = newRec.getPrimaryKey();
        Assert.assert(oldRec.getPrimaryKey().equals(primaryKey));
        try {
            getConnection().performUpdate(newRec);
        }
        catch (SQLException se) {
            throw new DataException(newRec, se);
        }
        newRec.touch();
        // The record is initialized, and hence, validated and immutable.
        freeze(newRec);
        markStale(oldRec);
        uncacheRec(oldRec);
        cacheRec(newRec);
        fireDataEvent(new DataEvent(this,
                                    DataEvent.RECORD_UPDATED,
                                    newRec,
                                    oldRec));
    }

    public List keys(String type) throws IOException {
        if (!Util.hasContent(type))
            type = defaultRecordType;
        if (completelyCachedTypes.contains(type)) synchronized(this) {
            ArrayList result = new ArrayList();
            String start = type + ":";
            int len = start.length();
            for (Iterator it = cache.keySet().iterator(); it.hasNext();) {
                String key = (String) it.next();
                if (key.startsWith(start))
                    result.add(key.substring(len));
            }
            return result;
        }
        try {
            return getConnection().getPrimaryKeys(type);
        }
        catch (SQLException se) {
            throw new DataException(null, se);
        }
    }

    public Record get(String type, Object key) throws IOException {
        if (type == null)
            type = defaultRecordType;
        Record result = (Record) getFromCache(type, key);
        if (result != null)
            return result;
        if (!completelyCachedTypes.contains(type)) {
            try {
                result = getConnection().retrieveValues(type, key);
            }
            catch (SQLException se) {
                throw new DataException(result, se);
            }
            if (result != null) {
                freeze(result);
                cacheRec(result);
            }
        }
        if (result == null) {
            String normkey = normalizeKey(type, key);
            if (!normkey.equals(key))
                return get(type, normkey);
        }
        return result;
    }

    synchronized Record getFromCache(String type, Object key) {
        return(Record) cache.get(type + ":" + key);
    }

    synchronized Record cacheRec(Record rec) {
        if (maxCacheSize <=0)
            return rec;
        if (cache.size() >= maxCacheSize)
            wipeCache();
        String lookupKey = rec.getType() + ":" + rec.getPrimaryKey();
        Record oldRec = (Record) cache.get(lookupKey);
        if (oldRec != null) {
            if (oldRec.equals(rec))
                return oldRec;
            else
                markStale(oldRec);
        }
        cache.put(lookupKey, rec);
        return rec;
    }

    synchronized void uncacheRec(Record rec) {
        cache.remove(rec.getType() + ":" + rec.getPrimaryKey());
    }

    synchronized public void wipeCache() {
        Map oldCache = this.cache;
        cache = new HashMap();
        for (Iterator it = oldCache.values().iterator(); it.hasNext();) {
            Record rec = (Record) it.next();
            markStale(rec);
        }
        completelyCachedTypes = new ArrayList();
    }

    // REVISIT: look more generally into the problem of
    // caching the results of the select queries.

    public List select(String type, RecordFilter filter) throws IOException {
        if (!Util.hasContent(type))
            type = defaultRecordType;
        List result = null;
        if (completelyCachedTypes.contains(type)) synchronized(this) {
            result = new ArrayList();
            for (Iterator it= keys(type).iterator(); it.hasNext();) {
                Record rec = (Record) get(type, it.next());
                if (filter == null || filter.accept(rec));
                    result.add(rec);
            }
            return result;
        }
        else {
            try {
                result = getConnection().selectRecords(type, filter);
            }
            catch (SQLException se) {
                throw new DataException(null, se);
            }
            int listSize = result.size();
            boolean cacheRecs = listSize < maxCacheSize/2; // INTERESTING PROBLEM: REVISIT (JR)
            if (filter == null && cacheRecs)
                completelyCachedTypes.add(type);
            for (int i = 0; i<listSize; i++) {
                Record rec = (Record) result.get(i);
                freeze(rec);
                if (cacheRecs) {
                    result.set(i, cacheRec(rec));
                }
            }
        }
        return result;
    }

    // TODO: Implement the possibility of a connection pool.
    // for efficiency sake (JR)
    protected JDBCConnection getConnection() {
        try {
            Assert.assert(!jdbcConnection.isClosed());
        }
        catch (Exception e) {
            jdbcConnection = new JDBCConnection(jdbcURL);
        }
        return jdbcConnection;
    }

    protected Object uniqueKey(Record rec) throws IOException {
        FieldDescriptor fd = rec.getMetadata().getPrimaryKeyField();
        boolean isInt = (fd.getJavaType() == Integer.class);
        int nextInt = Math.abs(rnd.nextInt());
        while (true) {
            Object next = new Integer(nextInt);
            if (!isInt) {
                next = next.toString();
            }
            if (get(rec.getType(), next) == null) {
                return next;
            }
        }
    }

    private Random rnd = new Random(); // used to generate keys
    private HashMap cache = new HashMap();
    private JDBCConnection jdbcConnection;
}
