/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.data.jdbc;

import java.io.*;
import java.sql.*;
import java.util.*;
import com.revusky.niggle.data.*;
import com.revusky.niggle.util.*;

/**
 * A wrapper object that wraps the connection to
 * an RDBMS. It also contains some
 * JDBC/SQL-related convenience routines
 * that bridge Niggle semantics with JDBC semantics.
 */

public class JDBCConnection {

    private Connection conn;

    public JDBCConnection(String jdbcURL) {
        try {
            this.conn = DriverManager.getConnection(jdbcURL);
        }
        catch (Exception e) {
            Assert.rethrow(e);
        }
    }

    public boolean isClosed() {
        try {
            return conn.isClosed();
        }
        catch (Exception e) {
            return true;
        }
    }

    public void close() {
        try {
            conn.close();
        }
        catch (Exception e) {
        }
        conn = null;
    }

    public void performDeletion(Record rec) throws SQLException {
        Statement stmt = conn.createStatement();
        RecordDescriptor metadata = rec.getMetadata();
        String query = "DELETE FROM "
                       + metadata.getTableName()
                       + " WHERE "
                       + metadata.getPrimaryKeyField().getName()
                       + "='"
                       + rec.getPrimaryKey()
                       + "'";
        int count = stmt.executeUpdate(query);
        Assert.assert(count != 0, "expected to find a record to delete in the RDBMS table!");
        if (count >1) {
            Logger.log("This is supposed to be a primary key and hence, have at most one record with a given field value.",
                       Logger.LOG_WARN);
        }
        stmt.close();
    }

    public void performInsertion(Record rec) throws SQLException {
        Assert.assert(rec.isImmutable());
        int count = 0;
        String query = SQLUtil.buildInsertQuery(rec);
        Statement stmt = conn.createStatement();
        count = stmt.executeUpdate(query);
        Assert.assert(count == 1, "We were supposed to have inserted a row here.");
        stmt.close();
    }

    public void performUpdate(Record rec) throws SQLException {
        String query = SQLUtil.buildUpdateQuery(rec);
        Statement stmt = conn.createStatement();
        int count = stmt.executeUpdate(query);
        Assert.assert(count == 1, "We were supposed to have updated one row here.");
        stmt.close();
    }

    public Record retrieveValues(String type, Object primaryKey) throws SQLException {
        RecordDescriptor metadata = null;
        try {
            metadata = DataRegistryImpl.getDefaultRegistry().getMetadata(type);
        }
        catch (Exception e) {
        } //We just end up with a NullPointerException bubbling up...
        String query = "SELECT * FROM "
                       + metadata.getTableName()
                       + " WHERE "
                       + metadata.getPrimaryKeyField().getName()
                       + "="
                       + SQLUtil.quoteValue(primaryKey);
        Statement stmt = conn.createStatement();
        ResultSet rs = stmt.executeQuery(query);
        if (!rs.next())
            return null;
        Record rec = metadata.newRecord();
        SQLUtil.fillRecordFromResultSet(rec, rs);
        stmt.close();
        return rec;
    }

    /**
     * Wraps the Statement.executeQuery
     * @param recordType The type of records in the list
     * @param postFilter Get rid of records that do not meet a certain criterion.
     * @return a List of niggle records of the given recordType
     */
    public List selectRecords(String recordType, RecordFilter filter)
    throws SQLException, IOException {
        if (filter == null)
            filter = RecordFilter.NULL_FILTER;
        boolean postFilter = false;
        String whereClause = "";
        try {
            whereClause = filter.sqlEquivalent();
        }
        catch (UnsupportedOperationException e) {
            postFilter = true;
        }
        RecordDescriptor metadata = DataRegistryImpl.getDefaultRegistry().getMetadata(recordType);
        String query = "SELECT * FROM " + metadata.getTableName() + whereClause;
        Statement stmt = conn.createStatement();
        ResultSet rs = stmt.executeQuery(query);
        List result = new ArrayList();
        while (rs.next()) {
            Record rec = metadata.newRecord();
            SQLUtil.fillRecordFromResultSet(rec, rs);
            if (filter == null || filter.accept(rec))
                result.add(rec);
        }
        stmt.close();
        return result;
    }


    public List getPrimaryKeys(String type) throws IOException, SQLException {
        List result = new ArrayList();
        DataRegistry reg = DataRegistryImpl.getDefaultRegistry();
        RecordDescriptor metadata = reg.getMetadata(type);
        String tableName = metadata.getTableName();
        FieldDescriptor primaryKeyField = metadata.getPrimaryKeyField();
        String primaryKeyName = primaryKeyField.getName();
        String query = "SELECT "
                       + primaryKeyName
                       + " FROM "
                       + type;
        Statement stmt = conn.createStatement();
        ResultSet rs = stmt.executeQuery(query);
        while (rs.next()) {
            //REVISIT: I don't know if getObject() is sufficient
            // generally. The primary key is usually a String or an int, I suppose. (JR)
            Object key = rs.getObject(primaryKeyName);
            key = primaryKeyField.convertFromJDBC(key);
            result.add(key);
        }
        stmt.close();
        return result;
    }
}
