/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.data.metadata;

import org.w3c.dom.*;
import org.xml.sax.SAXException;
import com.revusky.niggle.data.*;
import com.revusky.niggle.util.*;
import java.io.*;
import java.util.*;

/**
 * An object that provides metadata about how to instantiate
 * a DataSource object -- it bridges the W3C DOM with our
 * DataSource descriptions in XML format.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
public class DataSourceMetadata extends Properties {

    Class cl;
    DataSource instance;
    boolean remote;
    boolean preload;

    DataSourceMetadata() {
    }

    public void init(Element elem, ClassLoader cloader) throws SAXException
    {
        String classname = elem.getAttribute("CLASS");
        if (Util.hasContent(classname)) {
            setProperty("CLASS", classname);
        }
        String name = elem.getAttribute("NAME");
        if (!Util.hasContent(name)) {
            throw new SAXException("DATASOURCE must have a NAME attribute");
        }
        setProperty("NAME", name);
        this.remote = DOMUtil.isAffirmative(elem.getAttribute("REMOTE"));
        this.preload = !DOMUtil.isNegative(elem.getAttribute("PRELOAD"));
        NodeList props = elem.getElementsByTagName("PROPERTY");
        for (int i = 0; i<props.getLength(); i++) {
            Element e = (Element) props.item(i);
            String key = e.getAttribute("KEY");
            String value = e.getAttribute("VALUE");
            if (!Util.hasContent(key))
                throw new SAXException("Must specify KEY attribute for PROPERTY in DATASOURCE: " + name);
            if (value == null)
                throw new SAXException("Must specify VALUE attribute for PROPERTY " + key + " in DATASOURCE: " + name);
            if (getProperty(key) != null)
                throw new SAXException("KEY attribute " + key + " can only be specified once in DATASOURCE: " + name);
            setProperty(key, value);
        }
        String className = getProperty("CLASS");
        if (remote && getProperty("REMOTE_URL") == null)
            throw new SAXException("DATASOURCE " + name + " is specied as remote, but no REMOTE_URL property was specified.");
        else try {
                this.cl = Util.findClass(className, cloader);
            }
            catch (Exception e) {
                Logger.logException(e, Logger.LOG_ERROR);
            }
    }

    public DataSource getInstance() throws IOException
    {
        if (remote) {
            try {
                instance = (DataSource) java.rmi.Naming.lookup(getProperty("REMOTE_URL"));
            }
            catch (java.rmi.NotBoundException nbe) {
                throw new AssertionFailedException(nbe);
            }
        }
        else if (instance == null) {
            try {
                instance = (DataSource) cl.newInstance();
                instance.init(this);
            }
            catch (Exception e) {
                Logger.logException(e, Logger.LOG_ERROR);
            }
        }
        return instance;
    }

    public String getName() {
        return getProperty("NAME");
    }

    public boolean preloads() {
        return preload;
    }

    public boolean isRemote() {
        return remote;
    }

    /**
     * Checks for equality with another object.
     */
    public boolean equals(Object o) {
        if (super.equals(o))
            return true;
        if (o == null || !(o instanceof DataSourceMetadata))
            return false;
        return toString().equals(o.toString());
    }

    /**
     * outputs the key:value pairs that make up this metadata object.
     */
    public String toString() {
        StringBuffer result = new StringBuffer("DataSourceMetadata:\n");
        for (Enumeration e = keys(); e.hasMoreElements();) {
            Object key = e.nextElement();
            result.append(key.toString());
            result.append(": ");
            result.append(get(key));
            result.append("\n");
        }
        return result.toString();
    }
}
