/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.data.metadata;

import com.revusky.niggle.data.*;
import com.revusky.niggle.util.*;
import java.io.IOException;
import java.io.Serializable;
import java.util.Locale;
import java.util.StringTokenizer;
import org.w3c.dom.*;
import org.xml.sax.SAXException;


/**
 * abstract base class for objects that represent
 * the metadata relating to a record field. This object knows
 * how to set itself up from a DOM element.
 * @author <a href="mailto:jon@revusky.com>Jonathan Revusky</a>
 */
abstract public class Field implements FieldDescriptor {

    String name;
    private boolean required;
    protected String sqlType;
    protected String defaultValueString;
    transient private Object defaultValue;

    public Field() {
    }

    protected void init(Element element) throws SAXException {
        this.name = element.getAttribute("NAME");
        if (!Util.hasContent(name))
            throw new SAXException("A field must have a name specified");
        name = name.toLowerCase().intern();
        this.required = DOMUtil.isAffirmative(element.getAttribute("REQUIRED"));
        String sqlType = element.getAttribute("SQL_TYPE");
        if (Util.hasContent(sqlType))
            this.sqlType = sqlType;
        defaultValueString = element.getAttribute("DEFAULT");
        try {
            processDefaultValue();
        }
        catch (Exception e) {
            throw new SAXException(e);
        }
    }

    protected void processDefaultValue() throws IOException {
        if (Util.hasContent(defaultValueString)) {
            setDefaultValue(getObjectFromString(null, defaultValueString));
        }
    }

    public Object getCopy(Object o) {
        return o;
    }

    public String normalize(String s) {
        return s;
    }

    public final String getName() {
        return name;
    }

    public String toString() {
        return name;
    }

    public int hashCode() {
        return name.hashCode();
    }

    public boolean isRequired() {
        return required;
    }

    public String valueToString(Object val) {
        if (val == null)
            val = "";
        return val.toString();
    }

    public Class getJavaType() {
        return objectClass;
    }

    public String getSQLType() {
        if (sqlType == null)
            return "VARCHAR(254)"; // punt
        return sqlType;
    }

    public void checkValidValue(Record rec, Object o) throws IOException {
        if (o == null || "".equals(o)) {
            if (this.isRequired())
                throw new MangledDataException(rec,
                                               "Field " + name + " is required.");
            return;
        }

        if (!getJavaType().isInstance(o))
            throw new InvalidDataException(rec, "Invalid data for field " + name);
    }

    public final boolean isFieldOf(Record rec) {
        return rec.getMetadata().getField(name) == this;
    }

    public Object defaultValue() {
        if (defaultValue == null && Util.hasContent(defaultValueString)) {
            try {
                processDefaultValue();
            }
            catch (IOException ioe) {
                throw new AssertionFailedException(ioe);
            }
        }
        return defaultValue;
    }

    public String getLocalizedName(Locale locale) {
        StringTokenizer st = new StringTokenizer(name, "_", true);
        StringBuffer result = new StringBuffer();
        while (st.hasMoreTokens()) {
            String tok = st.nextToken();
            if (tok.equals("_"))
                result.append(" ");
            else
                result.append(capitalizeFirstLetter(tok));
        }
        return result.toString();
    }

    protected void setDefaultValue(Object o) {
        try {
            checkValidValue(null, o);
            this.defaultValue = o;
        }
        catch (IOException ioe) {
            Assert.rethrow(ioe);
        }
    }

    public static String capitalizeFirstLetter(String s) {
        int len = s.length();
        String firstChar = (len >0) ? s.substring(0, 1) : "";
        String trailing = (len > 1) ? s.substring(1) : "";
        return firstChar.toUpperCase() + trailing.toLowerCase();
    }

    public Object convertFromJDBC(Object o) {
        return o;
    }


    static final Class objectClass = Object.class;
}
