/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.data.metadata;

import com.revusky.niggle.data.*;
import com.revusky.niggle.util.*;
import org.w3c.dom.*;
import java.io.IOException;
import org.xml.sax.SAXException;

/**
 * A descriptor for a field that holds a single numerical value.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

public class NumericalField extends Field {

    private Object min, max;

    Class numericalType;

    public NumericalField() {
    }

    public void init(Element element) throws SAXException {
        String type = element.getAttribute("TYPE");
        type = Util.noNull(type).toLowerCase();
        if (type.equals("integer")) {
            this.numericalType = Integer.class;
            this.sqlType = "INTEGER";
        }
        else if (type.equals("long")) {
            this.numericalType = Long.class;
            this.sqlType = "BIGINT";
        }
        else if (type.equals("float")) {
            this.numericalType = Float.class;
            this.sqlType = "FLOAT";
        }
        else if (type.equals("double")) {
            this.numericalType = Double.class;
            this.sqlType = "DOUBLE";
        }
        else
            throw new SAXException("Must specify TYPE as: INTEGER, LONG, FLOAT or DOUBLE");
        NodeList mins = element.getElementsByTagName("MINIMUM");
        int len = mins.getLength();
        if (len  > 0) {
            if (len >1)
                throw new SAXException("The MINIMUM of a numerical field can only be specified once.");
            try {
                Element elem = (Element) mins.item(0);
                String value = elem.getAttribute("VALUE");
                min = number(value);
            }
            catch (NumberFormatException nfe) {
                throw new SAXException("Expected a number here.");
            }
        }
        NodeList maxes = element.getElementsByTagName("MAXIMUM");
        len = maxes.getLength();
        if (len  > 0) {
            if (len >1)
                throw new SAXException("The MAXIMUM of a numerical field can only be specified once.");
            try {
                Element elem = (Element) maxes.item(0);
                String value = elem.getAttribute("VALUE");
                max = number(value);
            }
            catch (NumberFormatException nfe) {
                throw new SAXException("Expected a number here.");
            }
        }
        if (min != null && max != null) {
            if (((Comparable) min).compareTo((Comparable) max) >=0)
                throw new SAXException("Minimum value must be less than maximum.");
        }
        super.init(element);
    }

    private Number number(String s) {
        if (numericalType == Integer.class)
            return new Integer(s);
        if (numericalType == Float.class)
            return new Float(s);
        if (numericalType == Long.class)
            return new Long(s);
        if (numericalType == Double.class)
            return new Double(s);
        throw new NumberFormatException();
    }

    public Class getJavaType() {
        return numericalType;
    }

    // REVISIT i18n of the error message

    public Object getObjectFromString(Record rec, String s) throws IOException {
        if (!Util.hasContent(s))
            return defaultValue();
        try {
            return number(s);
        }
        catch (NumberFormatException nfe) {
            String message = "Found " + s;
            if (numericalType == Float.class || numericalType == Double.class) {
                message += ": expected a floating point value here";
            }
            else message += ": expected an integer value here";
            throw new MangledDataException(rec, message);
        }
    }

    /**
     * Note that the min-max interval is [min, max)
     * following python range semantics.
     */
    public void checkValidValue(Record rec, Object value) throws IOException {
        super.checkValidValue(rec, value);
        if (value == null)
            return;
        if (min!=null) {
            if (((Comparable) min).compareTo(value) >0)
                throw new InvalidDataException(rec, "Value " + value
                                               + " is less than " + min);
        }
        if (max != null) {
            if (((Comparable) max).compareTo(value) <=0)
                throw new InvalidDataException(rec, "Value " + value
                                               + " is greater than " + max);
        }
    }
}
