/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.data.metadata;

import com.revusky.niggle.data.*;
import com.revusky.niggle.util.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * An implementation of the RecordDescriptor interface.
 * This class knows how to instantiate itself from a DOM Element
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
public class RecordMetadata implements RecordDescriptor {

    private Field[] fields;
    // Lookup for fieldname aliases.
    private HashMap aliases;
    private Class recordClass;
    private Class presentationWrapperClass;
    private Field primaryKeyField;
    transient private Record exemplar;
    private String type;
    private int version;
    private String tableName;


    public RecordMetadata() {
    }

    /**
     * @param the name of the field or an alias to the name
     */
    public FieldDescriptor getField(String name) {
        for (int i = 0; i<fields.length; i++) {
            Field field = fields[i];
            if (field.name == name)
                return field;
        }
        for (int i = 0; i<fields.length; i++) {
            Field field = fields[i];
            if (field.name.equals(name))
                return field;
        }
        if (aliases !=null)
            return(FieldDescriptor) aliases.get(name);
        return null;
    }

    public FieldDescriptor getField(int i) {
        return fields[i];
    }

    public Enumeration getFieldDescriptors() {
        return Util.arrayAsEnumeration(fields);
    }

    /**
     * fill in internal state from a DOM element
     * @param element A W3C DOM API object that represents
     * the XML fragment that we are processing.
     * @param recordData a dictionary of metadata objects that have
     * been processed so far.
     * @param cloader The servlet classloader if it is available, or null
     * @throws SAXException This implies that there is something invalid
     * in the XML.
     */

    protected void init(Element element, Map recordData, ClassLoader cloader) throws SAXException {
        String classname = element.getAttribute("CLASS");
        if (!Util.hasContent(classname))
            classname = "com.revusky.niggle.data.DefaultRecord";
        try {
            this.recordClass = Util.findClass(classname, cloader);
        }
        catch (ClassNotFoundException e) {
            throw new SAXException("Could not find class " + classname);
        }
        type = element.getAttribute("TYPE").intern();
        version = Integer.parseInt(element.getAttribute("VERSION"));
        String primaryKeyName = element.getAttribute("PRIMARY_KEY");
        processBaseRecords(element, recordData);
        processPresentationWrapper(element, cloader);
        processFields(element, cloader);
        if (Util.hasContent(primaryKeyName)) {
            this.primaryKeyField = (Field) getField(primaryKeyName);
            if (primaryKeyField == null) {
                throw new SAXException("primary key named: "
                                       + primaryKeyName
                                       + " is not a field of this record.");

            }
        }
        else
            this.primaryKeyField = fields[0];

//  This is probably not a bad thing to enforce, but I can't have it
//  during the transition away from the canonical primary key 'unique_id'
//  - PRM
//
//         if (!primaryKeyField.isRequired())
//             throw new SAXException("The primary key named "
//                                    +primaryKeyField.getName()
//                                    + " must be a required field.");
        String tableName = element.getAttribute("TABLE_NAME");
        if (Util.hasContent(tableName))
            this.tableName = tableName;

    }

    protected void processPresentationWrapper(Element element, ClassLoader cloader) throws SAXException
    {
        NodeList recs = element.getElementsByTagName("PRESENTATION");
        if (recs.getLength() == 1) {
            Element presentationElement = (Element) recs.item(0);
            String classname = presentationElement.getAttribute("CLASS");
            if (Util.hasContent(classname)) try {
                    presentationWrapperClass = Util.findClass(classname, cloader);
                }
                catch (Exception e) {
                    throw new SAXException(e);
                }
        }
        if (recs.getLength() > 1)
            throw new SAXException("The presentation class can only be specified once.");
    }

    /**
     * add the fields of the base record definition
     */

    private void processBaseRecords(Element element, Map recordData) throws SAXException {
        NodeList recs = element.getElementsByTagName("BASEREC");
        ArrayList baseFields = new ArrayList();
        for (int i = 0; i < recs.getLength() ; i++) {
            Element baseRecElement = (Element) recs.item(i);
            String id = baseRecElement.getAttribute("id");
            RecordMetadata baseRec = (RecordMetadata) recordData.get(id);
            if (baseRec == null) {
                throw new SAXException("No record found with id=" + id);
            }
            for (int j = 0; j<baseRec.fields.length; j++) {
                Field f = baseRec.fields[j];
                if (!baseFields.contains(f))
                    baseFields.add(f);
            }
        }
        fields = new Field[baseFields.size()];
        fields = (Field[]) baseFields.toArray(fields);
    }


    /**
      * Fills in the fields array based on the FIELD subelements
      * in the RECORD element passed in.
      */
    private void processFields(Element element, ClassLoader cloader) throws SAXException {
        NodeList fieldNodes = element.getElementsByTagName("FIELD");
        int numFields = fieldNodes.getLength();
        ArrayList fieldList = new ArrayList(numFields);

        for (int i=0; i<numFields; i++) {
            Element fieldElement = (Element) fieldNodes.item(i);
            String className = fieldElement.getAttribute("CLASS");
            try {
                Class cl = Util.findClass(className, cloader);
                Field field = (Field) cl.newInstance();
                field.init(fieldElement);
                String fieldName = field.getName();
                // Now check uniqueness of field name.
                for (Iterator it = fieldList.iterator(); it.hasNext();) {
                    Field f = (Field) it.next();
                    if (f.getName().equals(fieldName))
                        throw new SAXException("Field name " + fieldName
                                               + " was duplicated. Field names must be unique.");

                }
                fieldList.add(field);
                int j = fields.length;
                while (j>0) {
                    if (fields[--j].getName().equals(fieldName)) {
                        fieldList.remove(field);
                        fields[j] = field;
                    }
                }
                processAliases(fieldElement, field);
            }
            catch (Exception e) {
                if (e instanceof SAXException)
                    throw (SAXException) e;
                else // have to wrap it.
                    throw new SAXException(e);
            }
        }
        if (fields == null || fields.length == 0) {
            fields = new Field[fieldList.size()];
            fieldList.toArray(fields);
        }
        else {
            int numBaseFields = fields.length;
            int numNewFields = fieldList.size();
            Field[] ff = new Field[numBaseFields + numNewFields];
            System.arraycopy(fields, 0, ff, 0, numBaseFields);
            for (int i = 0; i<numNewFields; i++) {
                ff[numBaseFields + i] = (Field) fieldList.get(i);
            }
            this.fields = ff;
        }
    }

    public int size() {
        return fields.length;
    }

    /**
     * @return the index of the field of a given name, or -1 if it does not exist.
     * @param fieldname the name of the field or an alias
     */
    public int indexOf(String fieldname) {
        for (int i = 0; i < fields.length; i++) {
            if (fields[i].getName() == fieldname)
                return i;
        }
        for (int i = 0; i < fields.length; i++) {
            if (fields[i].getName().equals(fieldname))
                return i;
        }
        FieldDescriptor field = null;
        if (aliases != null) {
            field = (FieldDescriptor) aliases.get(fieldname);
        }
        if (field != null) {
            return indexOf(field.getName());
        }
        return -1;
    }

    private void processAliases(Element fieldElement, Field field) throws SAXException {
        NodeList aliasNodes = fieldElement.getElementsByTagName("ALIAS");
        for (int i = 0; i < aliasNodes.getLength(); i++) {
            Element alias = (Element) aliasNodes.item(i);
            String aliasName = DOMUtil.getText(alias);
            if (aliases != null && aliases.containsKey(aliasName)) {
                throw new SAXException("alias " + aliasName + " already defined.");
            }
            if (aliasName.equalsIgnoreCase(field.getName())) {
                throw new SAXException("alias " + aliasName + " must be different from the canonical field name.");
            }
            if (aliases == null)
                aliases = new HashMap();
            aliases.put(aliasName, field);
        }
    }

    public Class getPresentationWrapperClass() {
        return presentationWrapperClass;
    }

    public FieldDescriptor getPrimaryKeyField() {
        return primaryKeyField;
    }

    public Class getRecordClass() {
        return recordClass;
    }

    public Record newRecord() {
        if (exemplar == null) {
            synchronized (this) {
                try {
                    exemplar = (Record) recordClass.newInstance();
                }
                catch (IllegalAccessException iae) {
                    throw new AssertionFailedException("Problem instantiating class: "
                                                       + recordClass.getName()
                                                       + "\nNB. A record implementation "
                                                       + "must be public and have a "
                                                       + "public no-args constructor");

                }
                catch (InstantiationException ie) {
                    throw new AssertionFailedException("Problem instantiating class: "
                                                       + recordClass.getName()
                                                       + "\nNB. Cannot instantiate "
                                                       + " an abstract class or interface.");
                }
                exemplar.setMetadata(this);
            }
        }
        Record result = exemplar.getMutableCopy();
        result.clearFields();
        return result;
    }

    public String getType() {
        return type;
    }

    public int getVersion() {
        return version;
    }

    public String getTableName() {
        return(tableName == null) ? getType() : tableName;
    }
}
