/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.data.metadata;

import com.revusky.niggle.data.*;
import com.revusky.niggle.util.*;
import org.w3c.dom.*;
import gnu.regexp.*;
import java.util.*;
import java.io.IOException;
import java.util.StringTokenizer;
import org.xml.sax.SAXException;


/**
 * Field metadata that represents a field with a string value.
 * The string may be "normalized" to all caps, all lower case, etc.
 * Also, a regular expression pattern may be specified to validate
 * against.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
public class StringField extends Field {

    /**
     * A bitmask that determines how the string is "normalized".
     */
    int normalization;

    // The min and max lengths of the string.
    // Values of zero imply no minimum or maximum
    int minLength, maxLength;


    static final int LOWER_CASE = 1;
    static final int UPPER_CASE = 2;
    static final int CAPITALIZE = 4;
    static final int TRIM = 8;
    static final int TRIM_EXCESS_WHITESPACE = 16;

    // only relevant if maxLength is set.
    static final int TRUNCATE_SILENTLY = 32;

    // only relevant if minLength is set.
    static final int PAD_SILENTLY = 64;


    /**
     * The whitespace characters used when trimming
     * strings for normalization.
     */
    private String whitespacechars = " \t";

    /**
     * The regexp that expresses a constraint
     * that the field must satisfy.
     */
    private transient RE regexp;

    private String regexpString;

    public StringField() {
    }

    protected void init(Element element) throws SAXException {
        super.init(element);

        String min = element.getAttribute("MIN_LENGTH");
        String max = element.getAttribute("MAX_LENGTH");

        if (Util.hasContent(min)) try {
                minLength = Integer.parseInt(min);
                if (minLength <0)
                    throw new NumberFormatException();
            }
            catch (NumberFormatException nfe) {
                throw new SAXException("MIN_LENGTH must be a non-negative integer");
            }

        if (Util.hasContent(max)) try {
                maxLength = Integer.parseInt(max);
                if (maxLength <0)
                    throw new NumberFormatException();
            }
            catch (NumberFormatException nfe) {
                throw new SAXException("MAX_LENGTH must be a non-negative integer");
            }

        if (maxLength >0 && maxLength < minLength) {
            throw new SAXException("MAX_LENGTH cannot be less than MIN_LENGTH!");
        }


        NodeList normalize = element.getElementsByTagName("NORMALIZE");
        for (int i = 0; i<normalize.getLength(); i++) {
            processNormalization((Element) normalize.item(i));
        }
        NodeList expressions = element.getElementsByTagName("REGEXP");
        for (int i = 0; i<expressions.getLength(); i++) {
            processRegexp((Element) expressions.item(i));
        }

        if (!Util.hasContent(sqlType)) {
            if ((maxLength >0) && (maxLength <256))
                sqlType = "CHAR(" + maxLength + ")";
        }
    }

    public Class getJavaType() {
        return stringClass;
    }

    public String normalize(String s) {
        if ((normalization & LOWER_CASE) != 0)
            s = s.toLowerCase();
        if ((normalization & UPPER_CASE) != 0)
            s = s.toUpperCase();
        if ((normalization & CAPITALIZE) != 0) {
            s = capitalizeFirstLetter(s);
        }
        if ((normalization & TRIM) != 0)
            s = s.trim();
        if ((normalization & TRIM_EXCESS_WHITESPACE) != 0) {
            StringTokenizer st = new StringTokenizer(s, whitespacechars);
            StringBuffer buf = new StringBuffer();
            while (st.hasMoreTokens()) {
                buf.append(st.nextToken());
                buf.append((char) ' ');
            }
            s = buf.toString().trim();
        }
        if ((maxLength >0) && (normalization & TRUNCATE_SILENTLY) !=0) {
            if (s.length() > maxLength)
                s = s.substring(0, maxLength);
        }
        if (minLength >0 && (normalization & PAD_SILENTLY) != 0) {
            int diff = minLength - s.length();
            if (diff > 0) {
                StringBuffer buf = new StringBuffer(s);
                for (int i = 0; i< diff; i++)
                    buf.append((char) ' ');
                s = buf.toString();
            }
        }
        return s;
    }

    public Object getObjectFromString(Record rec, String s) throws MangledDataException {
        if (s.indexOf('\\') >=0) {
            s = decodeMultiLineString(s);
        }
        return s;
    }

    public String valueToString(Object o) {
        String s = (String) o;
        if (s.indexOf('\\') >=0 || s.indexOf('\n') >=0 || s.indexOf('\r') >=0) {
            s = encodeMultiLineString(s);
        }
        return s;
    }

    public void checkValidValue(Record rec, Object value) throws IOException
    {
        super.checkValidValue(rec, value);
        String s = (String) value;
        if (!Util.hasContent(s)) // nothing to check.
            return;
        if (minLength > 0) {
            if (s.length() < minLength)
                throw new MangledDataException("String " + s + " should be at least " + minLength + " characters.");
        }
        if (minLength > 0) {
            if (s.length() > maxLength)
                throw new MangledDataException("String " + s + " should be at most  " + maxLength + " characters.");
        }
        if (regexp != null) {
            if (!regexp.isMatch(value)) {
                throw new MangledDataException(rec, "String " + value
                                               + " is not a valid value for field " + this.getName()
                                               + ". It does not match the pattern: " + regexp);
            }
        }
    }


    private void processNormalization(Element elem) {
        String type = elem.getAttribute("TYPE");
        if ("LOWER_CASE".equals(type))
            normalization |= LOWER_CASE;
        else if ("UPPER_CASE".equals(type))
            normalization |= UPPER_CASE;
        else if ("CAPITALIZE".equals(type))
            normalization |= CAPITALIZE;
        else if ("TRIM".equals(type))
            normalization |= TRIM;
        else if ("TRIM_EXCESS_WHITE_SPACE".equals(type)) {
            normalization |= TRIM_EXCESS_WHITESPACE;
            String whitespacechars = elem.getAttribute("WHITESPACECHARS");
            if (Util.hasContent(whitespacechars))
                this.whitespacechars = whitespacechars;
        }
        else if ("TRUNCATE_SILENTLY".equals(type)) {
            normalization |= TRUNCATE_SILENTLY;
        }
    }

    /**
     * Massage a null into an empty string.
     * This is a touch that tends to avoid
     * unnecessary (and semantically meaningless)
     * NullPointerExceptions
     */
    public Object getCopy(Object o) {
        if (o==null)
            return "";
        return o;
    }

    private void processRegexp(Element elem) throws SAXException {
        regexpString = elem.getAttribute("EXPRESSION");
        try {
            this.regexp = new RE(regexpString);
        }
        catch (Exception e) {
            throw new SAXException(e);
        }
    }

    /**
     * If the regexp string is defined, reconstitute the
     * regexp object (since gnu.regexp.RE is not serializable)
     */
    private void readObject(java.io.ObjectInputStream s)
    throws IOException, ClassNotFoundException
    {
        s.defaultReadObject();
        try {
            if (regexpString != null)
                this.regexp = new RE(regexpString);
        }
        catch (Exception e) {
            Assert.rethrow(e);
        }
    }

    protected String encodeMultiLineString(String input) {
        StringBuffer buf = new StringBuffer();
        StringTokenizer st = new StringTokenizer(input, "\r\n\\", true);
        String token="", prevToken;
        while ( st.hasMoreTokens() ) {
            prevToken = token;
            token = st.nextToken();
            if ( token.equals("\n") ) {
                if ( !prevToken.equals("\r") ) {
                    buf.append("\\n");
                }
            }
            else if ( token.equals("\r") ) {
                buf.append("\\n");
            }
            else if ( token.equals("\\") ) {
                buf.append("\\\\");
            }
            else {
                buf.append(token);
            }
        }
        return buf.toString();
    }

    /**
     * Deal with a string with embedded \n literals for newlines.
     */
    protected String decodeMultiLineString(String input) {
        if ( input.indexOf('\\') <=0 ) return input;
        StringBuffer buf = new StringBuffer();
        int i=0;
        while ( i<input.length() ) {
            char c = input.charAt(i++);
            if ( c != '\\' ) buf.append(c);
            else {
                if ( i < input.length() ) {
                    c = input.charAt(i++);
                }
                if ( c == 'n' )
                    buf.append('\n');
                else if ( c== '\\' )
                    buf.append('\\');
                else {
                    buf.append('\\');
                    buf.append(c);
                }
            }
        }
        return buf.toString();
    }

    static final Class stringClass = String.class;
}
