/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.servlet;

import java.io.IOException;
import javax.servlet.*;
import javax.servlet.http.*;
import com.revusky.niggle.util.*;

/**
 * The base Servlet class in the Niggle framework.
 * A Niggle-based Servlet will either be an instance of
 * this class or of a subclass.
 * Note that this class defines very little
 * functionality. NiggleServlet delegates to 2
 * objects, NiggleConfig and ServletInteraction
 * @see com.revusky.niggle.servlet.NiggleConfig
 * @see com.revusky.niggle.servlet.ServletInteraction
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
abstract public class NiggleServlet extends HttpServlet {
    /**
     * The object that contains our servlet config info.
     */
    protected NiggleConfig niggleConfig;

    /**
     *  Servlet initialization. This routine just hands
     *  off to the NiggleConfig object that really
     *  takes care of initializing the framework.
     *  @see com.revusky.niggle.servlet.NiggleConfig
     *  @exception ServletException thown if there is
     *  some problem initializing this servlet. Servlet will
     *  not be loaded.
     */
    public void init(ServletConfig config) throws ServletException {
        if (config instanceof NiggleConfig)
            this.niggleConfig = (NiggleConfig) config;
        else
            this.niggleConfig = new NiggleConfig(config, this);
        super.init(niggleConfig);
    }

    /**
     * By default, a Niggle servlet treats a GET and a POST
     * identically. So this method, in this base class,
     * simply invokes doPost()
     * @see #doPost
     */
    public void doGet(HttpServletRequest request,
                      HttpServletResponse response)
    throws ServletException, IOException {
        doPost(request, response);
    }

    /**
     * This method implements the top-level flow of
     * of control by which we handle a request. Essentially,
     * we create a ServletInteraction object of the appropriate
     * type (we rely on the niggleConfig object to do this.)
     * Then we invoke the dispatch() method which carries
     * out the logic implicit in the request.
     * At the end, we invoke interaction.output()
     * in order to output a page as a result of the request.
     */
    public void doPost(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
        try {
            ServletInteraction interaction = niggleConfig.newServletInteraction(request, response);
            interaction.dispatch();
            interaction.output();
        }
        catch (Exception e) {
            handleException(response, e);
        }
    }

    public String getServletInfo() {
        return niggleConfig.getServletName();
    }

    /**
     * redefine this method in a subclass if you want to
     * throw in some extra default handling.
     * (For example, our in-house apps actually email the stack traces
     * to our dev and qa aliases.)
     * The default implementation here simply rethrows the exception and
     * it bubbles up to the handler at the
     * next level (the servlet server) which does whatever
     * it does by default.
     */
    protected void handleException(HttpServletResponse response, Exception e)
    throws IOException, ServletException {
        if (e instanceof IOException)
            throw (IOException) e;
        if (e instanceof ServletException)
            throw (ServletException) e;
        else
            Assert.rethrow(e);
    }
}
