/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.servlet;

import com.revusky.niggle.data.DataException;
import com.revusky.niggle.templates.Page;
import com.revusky.niggle.util.*;
import java.io.IOException;
import java.net.URLEncoder;
import java.util.*;
import java.rmi.RemoteException;
import java.lang.reflect.*;
import javax.servlet.http.*;

/**
 *  An abstract class that encapsulates the various paths of
 *  execution involved in fulfilling a servlet request. What is
 *  here is meant to be generic to the "typical" web app.
 *  Application-specific functionality will be supplied
 *  by concrete subclasses.
 *  @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
abstract public class ServletInteraction {
    /*
     * constants associated with a ServletInteraction instance
     */
    public final HttpServletRequest request;

    /**
     * Our HTML response back to the user.
     */
    public final HttpServletResponse response;

    /**
     * The information that contains the niggle servlet config info.
     */
    public final NiggleConfig config;

    /**
     * Our session, assuming we have one, and we are using
     * the servlet womb's default session-tracking scheme.
     */
    protected HttpSession session;

    /**
     * The session ID. If this is unset, we have no session.
     */
    protected String sessionID;

    /**
     * Login ID of the current user.
     */
    protected String userID;

    /**
      * locale associated with this request
      */
    protected Locale locale;

    /**
     * name of the current action being executed
     * by this ServletInteraction instance
     */
    protected String action;

    /**
     * The Page template object that we are working with.
     */
    protected Page page;

    /**
     * Indicates whether we handled the request by redirecting. A bit hacky.
     */
    protected boolean hasRedirected;

    public ServletInteraction(HttpServletRequest request,
                              HttpServletResponse response,
                              NiggleConfig config) 
    throws IOException {
        this.request = request;
        this.response = response;
        this.config = config;
        deduceAction();
        deduceSessionInfo();
        deduceLocale();
    }

    /**
     * A hook that you might want to override if you want
     * to be more sophisticated about how you decide
     * what the preferred locale for the request is.
     */
    protected void deduceLocale() {
        this.locale = request.getLocale();
    }

    /**
     * A hook that you might want to override if you want
     * to change the default scheme for how we decide
     * what the current <EM>action</EM> is.
     * If you override this, you will probably have to 
     * override the getURL() method also.
     * @see #getURL
     */
    protected void deduceAction() {
        this.action = getParameter(ACTION_KEY);
        if (!hasContent(action)) {
            action = "default";
        }
    }

    /**
     * A hook that you might want to override if you want
     * to change the default scheme for session tracking.
     * The default base implementation simply piggy-backs
     * on what the servlet runner provides. 
     */
    protected void deduceSessionInfo() {
        session = request.getSession(false);
       if (session != null) {
           sessionID = session.getId();
           userID = session.getAttribute(USER_ID_KEY).toString();
       }
    }

    /**
     * Override this if you want to change the default 
     * execXXX naming scheme. There is probably little 
     * practical reason to do this.
     */
    protected String methodNameFromAction(String action) {
        return EXEC_PREFIX
        + Character.toUpperCase(action.charAt(0))
        + action.substring(1);
    }

    /**
     * @return the preferred locale of this servlet request.
     */
    public final Locale getLocale() {
        return locale;
    }

    /**
     * Builds the URL that corresponds to a given action.
     * You might want to override this to automatically
     * put more information in a URL's query string.
     */
    public String getURL(String action) {
        String result = request.getRequestURI();
        if (hasContent(action) && !action.equals("default")) 
            result = result + '?' + ACTION_KEY + '=' + action;
        return result;
    }

    /**
     * Check for presence of data field in CGI input.
     * wrapper for the ServletRequest's getParameter method
     * returns "" instead of null if the param is not found,
     * avoids annoying glitches.
     */
    public String getParameter(final String s) {
        if ( !hasContent(s) )
            return "";
        String result = request.getParameter(s);
        return(result != null) ? result : "";
    }


    /**
     * @return whether the param is present in the servlet request.
     */
    public boolean hasParameter(String param) {
        return request.getParameter(param) != null;
    }

    /**
     * check for empty strings.
     * @return true if string at least one char (possibly blank).
     */
    public static boolean hasContent(String s) {
        return s != null && s.length() > 0;
    }

    /**
     * Finally output the page as a result of a servlet interaction
     */
    protected void output() throws IOException {
        if ( !hasRedirected ) {
            if ( page == null ) {
                response.sendError(response.SC_NOT_FOUND);
                return;
            }
            exposeDefaultVariables();
            page.outputPage(response, false);
        }
    }

    /**
     * Method that creates a new session.
     * This is a convenient method to override or extend
     * if you want certain things to happen right after 
     * a login, say.
     */
    protected boolean createNewSession() throws IOException {
        if (session != null) {
            session.invalidate();
            sessionID = null;
        }
        session = request.getSession(true);
        if (session != null) {
            sessionID = session.getId();
            if (hasContent(userID)) {
                session.setAttribute(USER_ID_KEY, userID);
            }
        }
        return hasSession();
    }

    /**
     * Method to call if a given action
     * requires a valid session.
     * NB: If there is no valid session info, but there is enough
     * info in the request to create a session
     * (typically id/password or possibly a cookie)
     * then create a new session. Otherwise, the user is unauthorized.
     *
     * @return true/false indicating whether the
     * login/session info is present. 
     */
    protected boolean checkSessionInfo() throws IOException {
        if (this.hasSession())
            return true;
        else if (hasValidLoginInfo()) {
            return createNewSession();
        }
        else {
            unauthorized();
            return false;
        }
    }

    /**
     *  A hook that can be overridden to make sure that
     * certain top-level template variables get exposed.
     */
    protected void exposeDefaultVariables() throws IOException {
        page.expose("_lang", locale.getLanguage());
        page.expose(SESSION_ID_KEY, sessionID);
        page.expose(USER_ID_KEY, userID);
        page.expose("servletPath", request.getServletPath());
        page.expose("requestURI", request.getRequestURI());
    }

    /**
     * default method to tell the user she's not authorized.
     * Subclass might override this and put up a login or signup 
     * page in order to be more graceful about things.
     */
    protected void unauthorized() throws IOException {
        response.sendError(401);
        hasRedirected = true;
    }

    /**
     * A convenient method to override if you want
     * to handle this kind of condition more gracefully.
     * The default base implementation sends a 404 error code.
     */
    protected void navigationError() throws IOException {
        response.sendError(404);
        hasRedirected = true;
    }

    /**
     * A default handler that simply displays the default template,
     * index.nhtml.
     */
    public void execDefault() throws IOException {
        try {
            page = getPage(DEFAULT_TEMPLATE);
        }
        catch (Exception e) {
            navigationError();
        }
    }

    /**
     * @return the Page object for a given template.
     * delegates to the PageFactory object.
     */
    protected Page getPage(String template) throws IOException {
        Page result = config.getPageFactory().getPage(template, this);
        if (result != null)
            result.expose("_template", template);
        return result;
    }

    /**
     * Indicates whether we have a session.
     */
    protected boolean hasSession() {
        return hasContent(sessionID);
    }

    /**
     * redirects the browser to the URL corresponding
     * to a given action.
     */
    protected void redirectAction(String action) throws IOException {
        String url = response.encodeRedirectURL(getURL(action));
        response.sendRedirect(url);
        this.hasRedirected = true;
    }

    /**
     * A convenience method to expose all of the
     * servlet request's parameters as template variables
     */
    protected void reexposeServletParams() {
        for ( Enumeration e = request.getParameterNames(); e.hasMoreElements(); ) {
            String param = (String) e.nextElement();
            page.expose(param, request.getParameter(param));
        }
    }

    protected void exposeActionURL(String action) {
        String url = getURL(action);
        page.expose(action + "URL", response.encodeURL(url));
    }

    /**
     * Override this method to check whether the servlet
     * request has the info for a login (thus allowing us
     * to create a new session). In this base implementation,
     * this always returns false. Subclasses will likely
     * override this.
     */
    protected boolean hasValidLoginInfo() throws IOException {
        return false;
    }

    /**
     * Dispatches the request based on the action
     * This actually delegates to the NiggleConfig
     * object that does some reflective magic
     * to invoke the method this.execFoo() for action "foo".
     */
    protected void dispatch() throws IOException {
        try {
            config.dispatchAction(this);
        }
        catch (DataException de) {
            de.setLocale(locale);
            recover(de);
        }
        catch (IOException e) {
            recover(e);
        }
    }

    /**
     * override this method to provide some default
     * exception handling for when an execXXX method
     * fails for whatever reason.
     * NB: This default implementation simply
     * rethrows the exception.
     */
    protected void recover(IOException e) throws IOException {
        throw e;
    }

    // constants 

    static final String ACTION_KEY = "action";
    static final String SESSION_ID_KEY="sid";
    static final String USER_ID_KEY = "uid";
    static final String EXEC_PREFIX = "exec";
    static final String DEFAULT_TEMPLATE = "index.nhtml";
}
