/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.templates.freemarkerimpl;

import java.util.*;
import freemarker.template.*;
import com.revusky.niggle.util.*;

/**
 * a freemarker template implementation that
 * wraps an enumeration of strings.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

public class EnumTemplateModel
implements TemplateListModel, TemplateScalarModel, TemplateHashModel {

    private int cursor = 0;
    private boolean[] selectedOptions;
    private String[] options;
    private Locale locale;
    private String title;
    private String dummyStartString;

    // only used if we have a dummyStartString
    private boolean afterDummy;

    public EnumTemplateModel(String[] options, String title, Locale locale) {
        this.options = options;
        this.title = title;
        this.locale = locale;
        this.selectedOptions = new boolean[options.length];
    }

    public EnumTemplateModel(String[] options, Locale locale) {
        this(options, null, locale);
    }

    /**
     *  Used to set a string that starts off the options.
     * This will very likely be something like "N/A"
     */
    public void setDummyStartString(String dummyStartString) {
        this.dummyStartString = dummyStartString;
    }

/**
 * Sets this string as a selected string.
 */
    public void setSelection(String selection) {
        for (int i=0; i<options.length; i++) {
            if (options[i].equalsIgnoreCase(selection)) {
                setSelection(i);
                break;
            }
        }
    }

/**
 * Sets the string at the specified index as selected
 * @throws ArrayIndexOutOfBoundsException
 */
    public void setSelection(int index) {
        selectedOptions[index] = true;
    }

    public String getAsString() {
        if (title!=null)
            return title;
        String result ="";
        for (int i=0; i<selectedOptions.length; i++) {
            if (selectedOptions[i]) {
                result = options[i]; // TODO: i18n
                break;
            }
        }
        return result;
    }

    public boolean isEmpty() {
        return false;
    }

    public boolean hasNext() {
        if (dummyStartString != null && !afterDummy)
            return true;
        return cursor < options.length;
    }

    public boolean isRewound() {
        return cursor == 0 && !afterDummy;
    }

    public void rewind() {
        afterDummy = false;
        cursor = 0;
    }

    public TemplateModel next() throws TemplateModelException {
        if (!hasNext()) throw new TemplateModelException();
        if (dummyStartString != null && !afterDummy) {
            afterDummy = true;
            return new DummyOption();
        }
        return new Option(cursor++);
    }

    public TemplateModel get(String key) throws TemplateModelException {
        for (int i = 0; i < options.length; i++) {
            if (key.equalsIgnoreCase(options[i]))
                return new Option(i);
        }
        return null;
    }

    public TemplateModel get(int i) throws TemplateModelException {
        if (i <0 || i>= options.length)
            throw new TemplateModelException("IndexOutOfBounds(" + i + ")");
        throw new UnsupportedOperationException();
    }

    public TemplateModel listSize() throws TemplateModelException {
        return new SmartScalar(new Integer(options.length).toString());
    }

    class Option implements TemplateHashModel, TemplateScalarModel {
        int index;

        Option(int index) {
            this.index = index;
        }

        public boolean isEmpty() {
            return false;
        }

        public TemplateModel get(String key) throws TemplateModelException {
            key = key.toLowerCase();
            if (key.equals("optiontag")) {
                String selected = isSelected() ? " SELECTED" : "";
                return new SmartScalar("<OPTION" +
                                       selected +
                                       " value=\"" +
                                       options[index] +
                                       "\">" + getAsString() + "</OPTION>");
            }
            else if (key.equals("checkboxtag")) {
                String checked = isSelected() ? "CHECKED " : " ";
                String s = "<input type=\"checkbox\" value=\"checked\""
                           + checked
                           + "name=\""
                           + options[index]
                           + "\">"
                           + getAsString()
                           + "</OPTION>";
                return new SmartScalar(s);
            }
            else if (key.equals("selected")) {
                return new SmartScalar(isSelected());
            }
            else return null;
        }

        // TODO: i18n

        public String getAsString() {
            return options[index];
        }

        private boolean isSelected() {
            return selectedOptions[index];
        }
    }

    class DummyOption implements TemplateHashModel, TemplateScalarModel {
        public String getAsString() {
            return dummyStartString;
        }

        public boolean isEmpty() {
            return false;
        }

        public TemplateModel get(String key) {
            if (key.equalsIgnoreCase("optiontag")) {
                return new SmartScalar("<OPTION" +
                                       " value=\"" +
                                       UNSET +
                                       "\">" + getAsString() + "</OPTION>");
            }
            else
                return null;
        }
    }
    public static final String UNSET = "";
}
