/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.templates.freemarkerimpl;

import com.revusky.niggle.servlet.NiggleConfig;
import com.revusky.niggle.templates.*;
import com.revusky.niggle.util.*;
import freemarker.template.*;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;


/**
 * An implementation of the com.revusky.niggle.templates.Page interface
 * that uses the freemarker template library.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
public class FreemarkerPage implements Page {

    protected LazyHash model;
    private Template pageTemplate;
    private String contentType = "text/html";
    private ClassLoader classLoader; // We need this to find ResourceBundles

    /**
     *  Creates a new FreemarkerPage.
     *  @param pageTemplate the template to use.
     *  @param model the model to use.
     *
     *  @param type the type of data to output.  May NOT be null.
     */
    public FreemarkerPage(Template pageTemplate, 
                          LazyHash model, 
                          ClassLoader classLoader,
                          String contentType) 
    throws IOException {
        this.pageTemplate = pageTemplate;
        this.classLoader = classLoader;
        this.contentType = contentType;
        if ( model == null )
            model = new LazyHash();
        this.model = model;
        model.setContentType(contentType);
    }

    /**
     *  Prepare the cooked HTML, filling in the parameters
     *  with data values, and send it to the end user
     *  via the servlet's response stream.
     *
     *  @param response handle to where the output should go.
     *  @exception IOException
     */
    public void outputPage(HttpServletResponse response, boolean isStatic) throws IOException {
        response.setContentType(contentType);
        if (!isStatic) {
            response.setHeader("Pragma", "no-cache"); // http 1.0
            response.setHeader("Cache-Control", "no-cache"); // http 1.1
        }
        PrintWriter out = response.getWriter();
        pageTemplate.process(model, out);
        out.close();
    }

    /**
     * @return the processed HTML as a String
     */
    public String toString() {
        StringWriter sw = new StringWriter();
        pageTemplate.process(model, new PrintWriter(sw));
        return sw.toString();
    }

    /**
     * expose internal data, publishing it to make it
     * available to a Freemarker Template.
     *
     * @param key    name of the field in the template
     * @param item   source of the data to populate the template
     */
    public void expose(String key, Object item) {
        model.put(key, item);
    }

    /**
     * Convenience method really, since all Freemarker
     * scalar variables are strings. Exposes the boolean
     * value as "true" or "". (Note that the empty string
     * in freemarker is essentially <em>false</em>
     *
     * @param key name of the field in the template
     * @param b sets to boolean value according to: b ? "true" : ""
     */
    public void expose(String key, boolean b) {
        model.put(key, b);
    }

    public void exposeResourceBundle(String lookupName) {
        ResourceBundle rb = null;
        Locale l = getLocale();
        if (classLoader != null) {
            rb = ResourceBundle.getBundle(lookupName, l, classLoader);
        }
        else {
            rb = ResourceBundle.getBundle(lookupName, l);
        }
        String key = lookupName.replace('.', '_');
        model.put(key, new ResourceBundleHash(rb));
    }

    // REVISIT!!!

    protected Locale getLocale() {
        if (pageTemplate instanceof LocalizedTemplate) {
            return ((LocalizedTemplate) pageTemplate).getLocale();
        }
        return model.getLocale();
    }
}
