/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.templates.freemarkerimpl;

import freemarker.template.*;
import java.util.*;
import com.revusky.niggle.data.*;
import com.revusky.niggle.servlet.ServletInteraction;
import com.revusky.niggle.util.*;

/**
 * A freemarker template model implementation
 * that replaces freemarker.template.SimpleHash.
 * This class adds the following:
 * (i) using this, the page template syntax is case-insensitive
 * (meshes better with HTML, which is case-insensitive)
 * (ii) the data in a niggle record can be exposed to the page in one go.
 * (This is a big deal!)
 * (iii) the data model encapsulates an i18n locale.
 * (This is a very big deal. This will be the main point at which the canonical
 * storage strings in a data record get internationalized.)
 * (iv) You can expose a nested List or Map in one go via put
 * (v) This uses lazy evaluation, so should be significantly more efficient!
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
public class LazyHash extends FreemarkerPresentationWrapper
implements TemplateModelRoot {

    HashMap hash = new HashMap();

    public LazyHash() {
    }

    public LazyHash(ServletInteraction interaction) {
        setSessionContext(interaction);
    }


    /**
     * Clears out the hash
     */
    public void clearData() {
        hash.clear();
    }

    /**
     * expose an arbitrary object in this hash model
     * @param the variable name by which to expose the records.
     * @param val the object to wrap as a freemarker template model.
     * expose on to a page: typically niggle records or scalars or lists
     * or nested hashes
     */
    public void put(String key, Object val) {
        if (CASE_INSENSITIVE) {
            key = key.toLowerCase();
        }
        hash.put(key, val);
    }

    /**
     * Overloaded so that the template variables
     * can be case-insensitive.
     */
    public void put(String key, boolean b) {
        if (CASE_INSENSITIVE) {
            key = key.toLowerCase();
        }
        hash.put(key, new SmartScalar(b));
    }

    public void put(String key, TemplateModel val) {
        if (CASE_INSENSITIVE) {
            key = key.toLowerCase();
        }
        hash.put(key, val);
    }

    /**
     * Overloaded so that the template variables
     * can be case-insensitive.
     */
    public TemplateModel get(String key) throws TemplateModelException {
        if (CASE_INSENSITIVE) {
            key = key.toLowerCase();
        }
        Object val = hash.get(key);
        TemplateModel result = createWrapper(val);
        if (val != result)
            hash.put(key, result);
        return(TemplateModel) result;
    }

    /**
     * Overloaded so that the template variables
     * can be case-insensitive.
     */
    public void remove(String key) {
        if (CASE_INSENSITIVE) {
            key = key.toLowerCase();
        }
        hash.remove(key);
    }

    /**
     * expose data from a record by key
     * @param key
     * @param rec source of data.
     */
    public void put(String key, Record rec) {
        put(key, createRecordWrapper(rec, interaction, contentType));
    }

    public boolean isEmpty() throws TemplateModelException {
        return hash.isEmpty();
    }
}
