/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.util;

import java.io.*;

/**
 *
 *  A helper class that gives us functionality equivalent to C's stdlib
 *  assert() macro.  Note that assertions should be guarded within code
 *  that is conditionally compiled.  Try to keep to Eiffel's assertion
 *  system.  Use static variables defined in a globally used interface
 *  class to control the inclusion/exclusion of the blocks of code that
 *  contain our assertions.  Fie on Joy & Gosling for not having built
 *  this into the language.
 * @author Paul Murphy murphy@niggle.com
 */
public class Assert {
    static private boolean fatal = false;

    private static void fail(String msg) {
        if (fatal) {
            new Exception().printStackTrace();
            System.exit(-1);
        }

        throw new AssertionFailedException(msg);
    }

    /**
     * Set fatal to true to cause system exit (JR)
     */

    public static void setFatal(boolean b) {
        Assert.fatal = b;
    }

    /**
     * The most basic assertion test - a boolean test with a message explaining it.
     */
    public static void assert(boolean b, String msg) {
        if (!b)
            fail(msg);
    }

    public static void assert(boolean b) {
        if (!b)
            fail(null);
    }

    /**
     * Special assertion for Strings. Checks for a string
     * being non-null AND non-zero length. (JR)
     */

    public static void hasContent(String s) {
        assert((s != null) && (s.length() != 0));
    }

    /**
     * convenience method to rethrow an exception
     * as a RuntimeException. t is an instanceof
     * RuntimeException, it just rethrows, otherwise,
     * it throws a wrapped AssertionFailedException
     */

    public static void rethrow(Throwable t) {
        if (t instanceof RuntimeException) {
            throw (RuntimeException) t;
        }
        throw new AssertionFailedException(t);
    }


    /**
     * convenient method that throws an IOException
     * if the file cannot be written to. To get through this method,
     * either (a) the file exists and we can write to it
     * or (b) the file does not exist, but the directory exists
     * and can be written to, i.e. we can create the file there.
     * And otherwise, we get an exception.
     */
    static public void canWriteToFile(String filename) throws IOException {
        File f = new File(filename);
        if (f.exists()) {
            if (!f.canWrite())
                throw new IOException("cannot write to file: " + f.getAbsolutePath());
        }
        else {
            File parent = f.getParentFile();
            if (parent == null || !parent.exists()) {
                throw new FileNotFoundException("No directory: " + parent);
            }
            if (!parent.canWrite()) {
                throw new IOException("cannot write to directory: " + parent.getAbsolutePath());
            }
        }
    }
}
