/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.util;

import java.io.*;
import java.text.*;
import java.util.*;
import org.xml.sax.SAXException;
import javax.servlet.ServletContext;

/**
 *  A class that manages logging.  Everything is setup and works through
 *  static methods.
 *  <p>
 *  Nothing needs to be done to use the logging system.  In that case,
 *  errors are logged to stderr.  By default, all log messages are logged.
 *  <p>
 *  A log file can be specified (<code>setLogFileName()</code>) and
 *  a log level can be set (<code>setLogLevel()</code>) in order
 *  to filter out less important  messages.  Log levels, in increasing
 *  order of severity are:  <code>LOG_DEBUG, LOG_INFO, LOG_WARN, and
 *  LOG_ERROR</code>.  <code>LOG_ERROR</code> cannot be filtered out.
 *  <p>
 *  The logging system can be setup from anywhere in a program, but
 *  it usually done from within <code>init()</code> or <code>main()</code>.
 * <p>
 * In the end an implementor of Log is actually used to perform the
 * logging. Any implementor of Log can be used.
 * <P>
 * I don't like this logging class much. It has no notion of the
 * Servlet or ServletContext from which you are logging.(JR)
 * </P>
 */
public class Logger {
    /*
     *  Log levels.
     */

    /**
     *  Log level.
     */
    public static final int LOG_ALL = 0;

    /**
     *  Log level.
     */
    public static final int LOG_DEBUG = 1;

    /**
     *  Log level.
     */
    public static final int LOG_INFO = 2;

    /**
     *  Log level.
     */
    public static final int LOG_WARN = 3;

    /**
     *  Log level.
     */
    public static final int LOG_ERROR = 4;

    /**
     * A Log implementor that echoes messages
     * to stderr.
     */

    static private Log _cons = new ConsoleLog(System.err);

    /**
     * The Log implementor that actually handles the output.
     * By default, this just echoes to stderr.
     */
    static private Log _log = _cons;

    /**
     *  The log level, between <code>LOG_ALL</code> (most verbose) to
     *  <code>LOG_ERROR</code>.
     */
    static private int _logLevel = LOG_ALL;

    static private boolean echoToCons = true;

    /**
     *  Set the log file name.
     *
     *  @param newFile the name of the log file to be used, which can be
     *  changed at any time.
     * Note that, by default, log messages will also to stderr.
     */
    static public void setLogFileName(String newFile) {
        setLogFileName(newFile, false);
    }

    static public void setServletContext(ServletContext servletContext) {
        if (servletContext != null) {
            setLog(new ServletLog(servletContext));
            Logger.echoToCons = false;
        }
    }

    /**
     *  Set the log file name.
     *
     *  @param newFile the name of the log file to be used, which can be
     *  changed at any time.
     * @param echoToCons indicates whether to echo log messages to stderr
     * as well.
     */
    static public void setLogFileName(String filename, boolean echoToCons) {
        try {
            Assert.canWriteToFile(filename);
            setLog(new FileLog(filename));
            Logger.echoToCons = echoToCons;
        }
        catch (Exception e) {
            setLog(new ConsoleLog());
            Logger.echoToCons = false;
        }
    }

    /**
     * Get the value of log.
     * @return Value of log.
     */
    static public Log getLog() {
        return _log;
    }

    /**
     * Set the value of log.
     * @param v  Value to assign to log.
     */
    static public void setLog(Log  v) {
        _log = v;
    }

    /**
     *  Set the log level.
     *
     *  @param newLevel the lowest level for which messages should be logged.
     *  <br><b>require:</b> <code>(newLevel == LOG_ALL) ||
     *  (newLevel == LOG_DEBUG) || (newLevel == LOG_INFO) ||
     *  (newLevel == LOG_WARN) || (newLevel == LOG_ERROR)</code>
     */
    static public void setLogLevel(int newLevel) {
        /*
         *  Make sure we're given a legal log level.
         */
        if ((newLevel < LOG_ALL) || (newLevel > LOG_ERROR)) {
            log("Illegal log level: " + String.valueOf(newLevel) + ".",
                LOG_ERROR);
            return;
        }

        _logLevel = newLevel;
    }

    /**
     *  Log the message.
     *
     *  @param msg the message to log.
     */
    static public void log(String msg) {
        log(msg, LOG_ALL);
    }

    /**
     *  Log the message at the specified level.
     *
     *  @param msg the message to log.
     *
     *  @param level the level at which to log the message.
     *  <br><b>require:</b> <code>
     *  (newLevel == LOG_DEBUG) || (newLevel == LOG_INFO) ||
     *  (newLevel == LOG_WARN) || (newLevel == LOG_ERROR)</code>
     */
    static public synchronized void log(String msg, int level) {
        /*
         *  Bail immediately if the message to be logged is at lower level
         *  than what is currently desired.
         */
        if (level < _logLevel)
            return;

        /*
         *  Prepare the string to be printed.
         */
        String ts_string = "";
        if (!(_log instanceof ServletLog)) // kludge
            ts_string = DATE_FORMAT.format(new Date());

        String level_string = null;
        if (level == LOG_DEBUG)
            level_string = "DEBUG";
        else if (level == LOG_INFO)
            level_string = "INFO";
        else if (level == LOG_WARN)
            level_string = "WARN";
        else // (level == LOG_ERROR)
            level_string = "ERROR";

        String logmsg = null;
        if (msg == null)
            logmsg = "(null)";
        else
            logmsg = msg;

        String logstring = null;
        if (level != LOG_ALL) {
            logstring =
            ts_string + " (" + level_string + "): " + logmsg;
        }
        else {
            logstring = ts_string + ": " + logmsg;
        }


        /*
         *  If the log file name is set, log to that file, if not, log
         *  to stderr.
         */
        Log log = getLog();
        log.println(logstring);
        if (echoToCons && log != _cons) {
            _cons.println(logstring);
        }
    }


    /**
     * log an exception, including the stack trace.
     * Also takes into account some exceptions that
     * "tunnel" other exceptions, such as AssertionFailedException
     * and org.xml.sax.SAXException
     * @author jrevusky
     */

    static public void logException(Throwable t, int level) {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        PrintStream ps = new PrintStream(baos);
        if (t instanceof AssertionFailedException) {
            AssertionFailedException afe = (AssertionFailedException) t;
            Throwable underlying = afe.getNested();
            ps.println("Assertion failed");
            if (underlying != null) {
                ps.println("Underlying exception: " + underlying.getClass().getName());
                t = underlying;
            }
        }
        if (t instanceof SAXException) {
            SAXException se = (SAXException) t;
            Exception underlying = se.getException();
            ps.println("Error parsing XML");
            if (underlying != null) {
                ps.println("Underlying exception: " + underlying.getClass().getName());
                t = underlying;
            }
        }
        t.printStackTrace(ps);
        log(baos.toString(), level);
    }

    static public void logException(Throwable t) {
        logException(t, LOG_ALL);
    }

    static private DateFormat DATE_FORMAT = DateFormat.getDateTimeInstance();
}
