/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.util;

import java.net.*;
import java.text.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import java.io.*;

/**
 *  Dumping ground for utility functions that are more or less general-purpose.
 */
public class Util {
    /**
     * An empty java.util.Enumeration, comes in handy
     * for API's that require an Enumeration to be passed in or out.
     */
    public static final Enumeration EMPTY_ENUM = new Enumeration() {
        public boolean hasMoreElements() {
            return false;
        }
        public Object nextElement() {
            throw new NoSuchElementException();
        }
    };

    /**
     * An empty java.util.Iterator, comes in handy
     * for API's that require an Iterator to be passed in or out.
     */
    public static final Iterator EMPTY_ITERATOR= new Iterator() {
        public boolean hasNext() {
            return false;
        }
        public Object next() {
            throw new NoSuchElementException();
        }
        public void remove() {
            throw new NoSuchElementException();
        }
    };

    /**
     * A zero integer wrapper, defined statically here to avoid replications
     */
    public static final Integer ZERO = new Integer(0);

    /**
     * @return the zero-length string if s is null,
     * otherwise just echoes
     */
    static public String noNull(String s) {
        return(s == null) ? "" : s;
    }

    /**
     *@return true if the string is null or zero-length
     */

    static public boolean hasContent(String s) {
        return(s != null) && (s.length() > 0);
    }

    /**
     * Gets the specified input stream, assuming that it
     * may be a fully specified filename, OR relative to the
     * system classpath.
     */
    static public InputStream getInputStream(String location) {
        return getInputStream(location, null);
    }

    static public InputStream getInputStream(String location, ClassLoader cl) {
        InputStream result = null;
        File f = new File(location);
        if (f.exists()) {
            try {
                return new FileInputStream(location);
            }
            catch (IOException e) {
            }
        }
        if (cl == null) {
            cl = Util.class.getClassLoader();
        }
        if (cl !=null)
            result = cl.getResourceAsStream(location);
        if (result == null)
            result = ClassLoader.getSystemResourceAsStream(location);
        return result;
    }

    static public Class findClass(String className, ClassLoader loader) 
    throws ClassNotFoundException {
        try {
            return Class.forName(className, true, loader);
        }
        catch (ClassNotFoundException e) {
        }
        ClassLoader loader2 = Util.class.getClassLoader();
        if (loader2 != loader) try {
                return Class.forName(className, true, loader2);
            }
            catch (ClassNotFoundException e2) {
            }
        if (loader != null && loader2!= null) { // See if we already tried this.
            return Class.forName(className, true, null);
        }
        throw new ClassNotFoundException("could not find class: " + className);
    }


    /**
     * wraps an array of objects as an instance of java.util.Enumeration
     * useful for API's that require an Enumeration to be passed in or out.
     */
    static public Enumeration arrayAsEnumeration(final Object[] arr) {
        return new Enumeration() {
            int index;
            public boolean hasMoreElements() {
                return index<arr.length;
            }

            public Object nextElement() {
                if ( !hasMoreElements() ) {
                    throw new NoSuchElementException();
                }
                return arr[index++];
            }
        };
    }

    /**
     * take a single object and wrap it up as an instance of java.util.Iterator
     * useful for API's that require an Iterator to be passed in or out.
     */
    static public Iterator objectAsIterator(final Object o) {
        return new Iterator() {
            boolean finished = false;

            public boolean hasNext() {
                return !finished;
            }

            public Object next() {
                if (finished)
                    throw new NoSuchElementException();
                finished = true;
                return o;
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    static public void printFullStackTrace(Throwable t, PrintStream stream) {
        while (t != null) {
            t.printStackTrace(stream);
            t = getNested(t);
            if (t != null)
                stream.println("Underlying cause:");
        }
    }

    static Throwable getNested(Throwable t) {
        if (t instanceof java.rmi.RemoteException) {
            java.rmi.RemoteException re = (java.rmi.RemoteException) t;
            return re.detail;
        }
        if (t instanceof org.xml.sax.SAXException) {
            org.xml.sax.SAXException se = (org.xml.sax.SAXException) t;
            return se.getException();
        }
        if (t instanceof com.revusky.niggle.data.DataException) {
            com.revusky.niggle.data.DataException de = (com.revusky.niggle.data.DataException) t;
            return de.getNestedException();
        }
        if (t instanceof AssertionFailedException) {
            AssertionFailedException ae = (AssertionFailedException) t;
            return ae.getNested();
        }
        return null;
    }


    /**
     * Converts the literal characters in a string
     * to HTML entities as need be.
     * @return converted string or simply echoes if no 
     * conversion was necessary.
     */
    static public String convertHTMLEntities(String s) {
        StringBuffer buf = new StringBuffer();
        boolean converted = false;
        for (int i = 0; i<s.length(); i++) {
            char c = s.charAt(i);
            int offset = LITERALS.indexOf(c);
            if (offset < 0) {
                buf.append(c);
            }
            else {
                buf.append(ENTITIES[offset]);
                converted = true;
            }
        }
        return converted ? buf.toString() : s;
    }

    /**
     * A routine that says whether the given string
     * has unicode characters in it -- by that I mean,
     * characters above 0x7F, for which a naive byte encoding
     * will not work.
     */
    static public boolean hasNonAscii(String s) {
        char[] chars = s.toCharArray();                    
        for (int i = 0; i< chars.length; i++) {
            if ((chars[i] & 0xFF80) != 0) {
                return true;
            }
        }
        return false;
    }

    /**
     * @return a String containing the UTF-8 encoding of 
     * the given string.
     */
    static public String toByteEncodedString(String s) {
        try {
            byte[] bytes = s.getBytes("UTF-8");
            return new String(bytes, "ISO-8859-1");
        }
        catch (Exception e) {
            Assert.rethrow(e);
            return null; //makes compiler happy
        }
    }

    /**
     * @return a String containing the straight unicode chars
     * from previous byte encoding.
     */

    static public String fromByteEncodedString(String s) {
        try {
            byte[] bytes = s.getBytes("ISO-8859-1");
            return new String(bytes, "UTF-8");
        }
        catch (Exception e) {
            Assert.rethrow(e);
            return null; //makes compiler happy
        }
    }

    static public String randomAlphaNumString(int length) {
        String result = "";
        while (result.length() < length) {
            int i = Math.abs(rnd.nextInt());
            result += Integer.toString(i, 36);
        }
        return result.substring(0, length);
    }






    // TODO: Add more literals and entities.

    static final String LITERALS  = "\"<>";

    static final String[] ENTITIES = new String[] {
        "&QUOT;",
        "&LT;",
        "&GT;"
    };

    static Random rnd = new Random();

}
